from __future__ import annotations
from typing import TYPE_CHECKING

from code import InteractiveConsole
from collections import deque
import asyncio
from functools import cache
from shlex import shlex
import sys
import re
from trueseeing.core.context import FileOpener
from trueseeing.core.ui import ui, CoreProgressReporter
from trueseeing.core.exc import FatalError, InvalidSchemaError

if TYPE_CHECKING:
  from typing import Mapping, Optional, Any, NoReturn, List, Dict, Awaitable, Type, Set, Iterator
  from prompt_toolkit.styles import Style
  from trueseeing.core.context import ContextType
  from trueseeing.api import Entry, Command, CommandHelper, CommandEntry, CommandPatternEntry, ModifierEntry, OptionEntry, ConfigMap, ModifierEvent, CommandMap, CommandPatternMap, OptionMap, ModifierMap

class InspectMode:
  def do(
      self,
      target: Optional[str],
      batch: bool = False,
      cmdlines: List[str] = [],
      abort_on_errors: bool = False,
      force_opener: Optional[str] = None,
  ) -> NoReturn:
    try:
      try:
        if ui.is_tty():
          ui.enter_inspect()
        with CoreProgressReporter().scoped():
          self._do(target, batch, cmdlines, abort_on_errors, force_opener)
      finally:
        if ui.is_tty():
          ui.exit_inspect()
    except QuitSession as e:
      sys.exit(e.code)
    else:
      sys.exit(0)

  def _do(
      self,
      target: Optional[str],
      batch: bool,
      cmdlines: List[str],
      abort_on_errors: bool,
      force_opener: Optional[str],
  ) -> int:
    runner = Runner(target, abort_on_errors=abort_on_errors, force_opener=force_opener)

    for line in cmdlines:
      asyncio.run(LambdaConsole._worker(runner.run(line)))

    if batch:
      return 0

    if not ui.is_tty(stdin=True):
      ui.fatal('requires a tty')

    asyncio.run(runner.greeting())

    ps1, ps2 = getattr(sys, 'ps1', None), getattr(sys, 'ps2', None)
    try:
      LambdaConsole(locals=locals(), runner=runner).interact(banner='', exitmsg='')
      return 0
    finally:
      sys.ps1, sys.ps2 = ps1, ps2

class LambdaConsole(InteractiveConsole):
  def __init__(self, /, runner: Runner, locals: Optional[Mapping[str, Any]] = None) -> None:
    super().__init__(locals=locals, filename='<input>')
    from prompt_toolkit import PromptSession
    self._sess: Any = PromptSession()
    self._runner = runner

  def runsource(self, source: str, filename: Optional[str]=None, symbol: Optional[str]=None) -> bool:
    try:
      asyncio.run(self._worker(self._runner.run(source)))
    except FatalError:
      pass
    return False

  def raw_input(self, prompt: str = "") -> str:
    target = self._runner.get_target()
    return self._sess.prompt(  # type: ignore[no-any-return]
      message=[('class:p', f'ts[{target}]> ' if target else 'ts> ')],
      prompt_continuation=[('class:p', '... ')],
      style=self._get_prompt_style(),
    )

  @cache
  def _get_prompt_style(self) -> Style:
    from prompt_toolkit.styles import Style
    return Style.from_dict({'p': '#888800'})

  @classmethod
  async def _worker(cls, coro: Any) -> None:
    tasks = [asyncio.create_task(coro)]
    done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
    for t in pending:
      t.cancel()
    if pending:
      _, _ = await asyncio.wait(pending)
    for t in done:
      if not t.cancelled():
        x = t.exception()
        if x:
          assert isinstance(x, Exception)
          if isinstance(x, QuitSession):
            raise x
          elif not isinstance(x, FatalError):
            ui.fatal('unhandled exception', exc=x)

class QuitSession(Exception):
  def __init__(self, code: int, *args: Any, **kwargs: Any) -> None:
    super().__init__(code, *args, **kwargs)
    self.code = code

class SessionCommands:
  _cmds: Dict[str, Dict[str, CommandEntry]]
  _cmdpats: Dict[str, Dict[str, CommandPatternEntry]]
  _mods: Dict[str, Dict[str, ModifierEntry]]
  _opts: Dict[str, Dict[str, OptionEntry]]
  _typ: Optional[Set[ContextType]] = None

  def __init__(self) -> None:
    from trueseeing.core.config import Configs
    self._cmds = {}
    self._cmdpats = {}
    self._mods = {}
    self._opts = {}
    self._confbag = Configs.get().bag

  def add_cmds(self, cmap: CommandMap) -> None:
    for k, v in cmap.items():
      self._cmds[k] = {p:v for p in v.get('t', {''})}

  def add_cmdpats(self, cmap: CommandPatternMap) -> None:
    for k, v in cmap.items():
      self._cmdpats[k] = {p:v for p in v.get('t', {''})}

  def add_mods(self, cmap: ModifierMap) -> None:
    for k, v in cmap.items():
      self._mods[k] = {p:v for p in v.get('t', {''})}

  def add_opts(self, cmap: OptionMap) -> None:
    for k, v in cmap.items():
      self._opts[k] = {p:v for p in v.get('t', {''})}

  def add_configs(self, cmap: ConfigMap) -> None:
    self._confbag.update(cmap)

  def get_cmds(self) -> CommandMap:
    return self._slice(self._cmds) # type:ignore[no-any-return]

  def get_cmdpats(self) -> CommandPatternMap:
    return self._slice(self._cmdpats) # type:ignore[no-any-return]

  def get_mods(self) -> ModifierMap:
    return self._slice(self._mods) # type:ignore[no-any-return]

  def get_opts(self) -> OptionMap:
    return self._slice(self._opts) # type:ignore[no-any-return]

  def get_configs(self) -> ConfigMap:
    return self._confbag

  def _slice(self, target: Any) -> Any:
    o: Dict[str, Any] = dict()
    for k,pv in target.items():
      for p in self._get_matches(pv.keys()):
        o[k] = pv[p]
        break
    return o

  def _get_matches(self, ks: Iterator[str]) -> Iterator[str]:
    for k in sorted(ks, key=len):
      if self._typ is None:
        yield k
      elif not k:
        yield k
      elif any([re.match(k, t) for t in self._typ]):
        yield k

  def set_type(self, typ: Optional[Set[ContextType]]) -> None:
    self._typ = typ

class Runner:
  _sc: SessionCommands
  _quiet: bool = False
  _verbose: bool = False
  _target: Optional[str]
  _abort_on_errors: bool = False
  _helper: CommandHelper
  _loglevel: int

  def __init__(self, target: Optional[str], *, abort_on_errors: bool = False, force_opener: Optional[str] = None) -> None:
    self._target = target
    self._sc = SessionCommands()
    self._sc.add_cmds({
      '?':dict(e=self._help, n='?', d='help'),
      '?@?':dict(e=self._help_mod, n='?@?', d='modifier help'),
      '?o?':dict(e=self._help_opt, n='?o?', d='options help'),
      '?f?':dict(e=self._help_formats, n='?f?', d='supported file formats'),
      '!':dict(e=self._shell, n='!', d='shell'),
      'o':dict(e=self._set_target, n='o /path/to/target', d='set target file'),
      'q':dict(e=self._quit, n='q', d='quit'),
    })
    self._sc.add_mods({
      'o':dict(n='@o:option', d='pass option', e=None),
      'gs':dict(n='@gs:<int>[kmKM]', d='set graph size limit', e=None),
    })
    self._sc.add_configs(self._get_configs())
    if abort_on_errors:
      self._abort_on_errors = True

    self._helper = CommandHelperImpl(self, force_opener=force_opener)
    self._loglevel = ui.level

    self._init_cmds()

  def _init_cmds(self) -> None:
    from trueseeing.core.ext import Extension
    from trueseeing.app.cmd import discover

    for clazz in discover():
      self._register_cmd(clazz)

    for clazz in Extension.get().get_commands():
      self._register_cmd(clazz)

  def _register_cmd(self, clazz: Type[Command]) -> None:
    t = clazz.create(self._helper)
    self._sc.add_cmds(t.get_commands())
    self._sc.add_cmdpats(t.get_command_patterns())
    self._sc.add_mods(t.get_modifiers())
    self._sc.add_opts(t.get_options())
    self._sc.add_configs(t.get_configs())

  def get_target(self) -> Optional[str]:
    return self._target

  def _get_configs(self) -> ConfigMap:
    return {
      'core.debug':dict(g=self._config_get_debug, s=self._config_set_debug, n='core.debug', d='toggle debug mode (true, false)'),
      'core.quiet':dict(g=self._config_get_quiet, s=self._config_set_quiet, n='core.quiet', d='toggle quiet mode (true, false)'),
    }

  def _config_get_debug(self) -> str:
    return 'true' if (ui.level == ui.DEBUG) else 'false'

  def _config_set_debug(self, v: Any) -> None:
    try:
      self._loglevel = dict(true=ui.DEBUG, false=ui.INFO)[v]
      self._reset()
    except KeyError:
      ui.fatal(f'invalid value: {v}')

  def _config_get_quiet(self) -> str:
    return 'true' if (ui.level == ui.WARN) else 'false'

  def _config_set_quiet(self, v: Any) -> None:
    try:
      self._loglevel = dict(true=ui.WARN, false=ui.INFO)[v]
    except KeyError:
      ui.fatal(f'invalid value: {v}')

  async def greeting(self) -> None:
    try:
      from trueseeing import __version__ as version
      ui.success(f"Trueseeing {version}")
    finally:
      self._reset()

  async def run(self, s: str) -> None:
    try:
      try:
        await self._run(s)
      except InvalidSchemaError:
        ui.fatal('invalid schema detected, forced reanalysis needed (try a!)')
    finally:
      self._reset()

  async def _run(self, s: str) -> None:
    if not await self._run_raw(s):
      o: deque[str] = deque()
      lex = shlex(s, posix=True, punctuation_chars=';')
      lex.wordchars += '@:,=!$'
      for t in lex:
        if re.fullmatch(';+', t):
          if not await self._run_cmd(o, line=None):
            ui.error('invalid command, type ? for help')
          o.clear()
        else:
          o.append(t)
      if o:
        if not await self._run_cmd(o, line=s):
          ui.error('invalid command, type ? for help')

  async def _run_raw(self, line: str) -> bool:
    ent: Any = None
    cmdpats = self._sc.get_cmdpats()
    for pat in [k for k,v in cmdpats.items() if v.get('raw')]:
      m = re.match(pat, line)
      if m:
        ent = cmdpats[pat]
        if m.end() < (len(line) - 1):
          ui.warn('ignoring trailer: {}'.format(line[m.end():]))
        break
    if ent is None:
      return False
    else:
      assert m is not None
      await self._as_cmd(ent['e'](line=m.group(0)))
      return True

  async def _run_cmd(self, tokens: deque[str], line: Optional[str]) -> bool:
    ent: Any = None
    cmdpats = self._sc.get_cmdpats()
    cmds = self._sc.get_cmds()
    if line is not None:
      for pat in [k for k,v in cmdpats.items() if not v.get('raw')]:
        m = re.match(pat, line)
        if m:
          ent = cmdpats[pat]
          break

    if ent is None:
      c = tokens[0]
      if c in cmds:
        ent = cmds[c]

    if ent is None:
      return False
    else:
      await self._notify_modifiers('begin', tokens)
      try:
        await self._as_cmd(ent['e'](args=tokens))
        return True
      finally:
        await self._notify_modifiers('end', tokens)

  async def _as_cmd(self, coro: Awaitable[Any]) -> None:
    try:
      try:
        await coro
      except KeyboardInterrupt:
        ui.fatal('interrupted')
    except FatalError:
      if self._abort_on_errors:
        raise QuitSession(1)

  async def _notify_modifiers(self, ev: ModifierEvent, tokens: deque[str]) -> None:
    for mod in self._get_modifiers(tokens):
      typ, val = mod[1:].split(':', maxsplit=1)

      for et, ee in self._sc.get_mods().items():
        if et == typ:
          if ee['e'] is not None:
            await ee['e'](ev, val)

  def _get_modifiers(self, args: deque[str]) -> List[str]:
    o = []
    for m in args:
      if m.startswith('@'):
        o.append(m)
    return o

  def _reset(self, debug:bool = False) -> None:
    ui.set_level(self._loglevel)
    self._sc.set_type(self._helper.get_context_type())

  async def _help(self, args: deque[str]) -> None:
    ents: Dict[str, Entry] = dict()
    ents.update(self._sc.get_cmds())
    ents.update(self._sc.get_cmdpats())  # type: ignore[arg-type]
    await self._help_on('Commands:', ents)

  async def _help_mod(self, args: deque[str]) -> None:
    await self._help_on('Modifiers:', self._sc.get_mods()) # type: ignore[arg-type]

  async def _help_opt(self, args: deque[str]) -> None:
    await self._help_on('Options:', self._sc.get_opts()) # type: ignore[arg-type]

  async def _help_on(self, topic: str, entries: Dict[str, Entry]) -> None:
    ui.success(topic)
    if entries:
      width = (2 + max([len(e.get('d', '')) for e in entries.values()]) // 4) * 4
      for k in sorted(entries):
        e = entries[k]
        if 'n' in e:
          ui.stderr(
            f'{{n:<{width}s}}{{d}}'.format(n=e['n'], d=e['d'])
          )

  async def _help_formats(self, args: deque[str]) -> None:
    from trueseeing.core.context import FileOpener
    ui.success('File formats:')
    formats = list(FileOpener().get_formats())
    width = 2 + max([len(e['n']) for e in formats])
    for e in formats:
      ui.stderr(
        f'{{n:<{width}s}}{{d}}'.format(n=e['n'], d=e['d'])
      )

  async def _shell(self, args: deque[str]) -> None:
    from trueseeing.core.env import get_shell
    from asyncio import create_subprocess_exec
    await (await create_subprocess_exec(get_shell())).wait()

  async def _quit(self, args: deque[str]) -> None:
    raise QuitSession(0)

  async def _set_target(self, args: deque[str]) -> None:
    _ = args.popleft()

    if not args:
      ui.fatal('need path')

    self._target = args.popleft()

class CommandHelperImpl:
  def __init__(self, runner: Runner, force_opener: Optional[str] = None) -> None:
    self._r = runner
    self._opener = FileOpener(force_opener=force_opener)
    self._confbag = self._r._sc.get_configs()

  def get_target(self) -> Optional[str]:
    return self._r.get_target()

  def require_target(self, msg: Optional[str] = None) -> str:
    t = self.get_target()
    if t is None:
      ui.fatal(msg if msg else 'need target')
    return t

  def get_context_type(self) -> Optional[Set[ContextType]]:
    from trueseeing.core.exc import InvalidFileFormatError
    t = self.get_target()
    if not t:
      return None
    try:
      return self._opener.get_context(t).type
    except InvalidFileFormatError:
      return set()

  def get_context(self, typ: Optional[ContextType] = None) -> Any:
    from trueseeing.core.exc import InvalidFileFormatError
    try:
      c = self._opener.get_context(self.require_target())
      if typ is not None:
        ui.warn('get_context(typ): deprecated, use get_context().request_type(...)', onetime=True)
        c.require_type(typ)
      return c
    except InvalidFileFormatError:
      ui.fatal('cannot recognize format')

  async def get_context_analyzed(self, typ: Optional[ContextType] = None, *, level: int = 3) -> Any:
    if typ is None:
      ui.warn('get_context_analyzed: deprecated; use get_context().analyze(...)', onetime=True)
    else:
      ui.warn('get_context_analyzed(typ): deprecated; use get_context().request_type(...).analyze(...)', onetime=True)

    c = self.get_context()
    if typ:
      c.require_type(typ)
    await c.analyze(level=level)
    return c

  def decode_analysis_level(self, level: int) -> str:
    analysislevelmap = {0:'no', 1: 'minimally', 2: 'lightly', 3:'marginally', 4:'fully'}
    return analysislevelmap.get(level, '?')

  async def run(self, s: str) -> None:
    await self._r._run(s)

  async def run_cmd(self, tokens: deque[str], line: Optional[str]) -> bool:
    return await self._r._run_cmd(tokens, line)

  def get_modifiers(self, args: deque[str]) -> List[str]:
    return self._r._get_modifiers(args)

  def get_effective_options(self, mods: List[str]) -> Mapping[str, str]:
    o = {}
    for m in mods:
      if m.startswith('@o:'):
        for a in m[3:].split(','):
          c: List[str] = a.split('=', maxsplit=1)
          if len(c) == 1:
            o[c[0]] = c[0]
          else:
            o[c[0]] = c[1]
    return o

  def get_graph_size_limit(self, mods: List[str]) -> Optional[int]:
    for m in mods:
      if m.startswith('@gs:'):
        c = m[4:]
        s = re.search(r'[km]$', c.lower())
        if s:
          return int(m[4:-1]) * dict(k=1024, m=1024*1024)[s.group(0)]
        else:
          return int(m[4:])
    return None

  def get_config(self, k: str) -> Any:
    from trueseeing.core.exc import InvalidConfigKeyError
    try:
      e = self._confbag[k]
    except KeyError:
      raise InvalidConfigKeyError()
    else:
      return e['g']()

  def set_config(self, k: str, v: Any) -> None:
    from trueseeing.core.exc import InvalidConfigKeyError
    try:
      e = self._confbag[k]
    except KeyError:
      raise InvalidConfigKeyError()
    else:
      e['s'](v)



from __future__ import annotations
from typing import TYPE_CHECKING

import math
import re
from trueseeing.api import Signature
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Dict, AnyStr
  from trueseeing.api import SignatureMap, SignatureHelper, ConfigMap
  from trueseeing.core.android.context import APKContext
  from trueseeing.core.android.model import Call

class NativeCodeDetector(Signature):
  _cvss_info = 'CVSS:3.0/AV:P/AC:L/PR:N/UI:N/S:U/C:N/I:N/A:N/'

  def __init__(self, helper: SignatureHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return NativeCodeDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {
      'nat-detect-api':dict(e=self._detect_api, d='Detects API call in native code'),
      'nat-detect-urls':dict(e=self._detect_url, d='Detects URL etc. in native code'),
    }

  def get_configs(self) -> ConfigMap:
    return dict()

  def _get_context(self) -> APKContext:
    return self._helper.get_context().require_type('apk')

  def _format_aff0(self, c: Call) -> str:
    return self._format_aff0_manual(c['path'], c['sect'], c['offs'])

  def _format_aff0_match(self, n: str, m: re.Match[AnyStr]) -> str:
    return self._format_aff0_manual(n, '', m.start())

  def _format_aff0_manual(self, n: str, s: str, o: int) -> str:
    return '{} ({}+{:08x})'.format(n, s, o)

  async def _detect_api(self) -> None:
    context = self._get_context()
    with context.store().query().scoped() as q:
      for c in q.calls():
        priv, target = c['priv'], c['target']
        self._helper.raise_issue(self._helper.build_issue(
          sigid='nat-detect-api',
          cvss=self._cvss_info,
          title='detected {} call'.format('private' if priv else 'API'),
          info0=target,
          aff0=self._format_aff0(c),
        ))

  async def _detect_url(self) -> None:
    from trueseeing.core.analyze import analyze_url_in
    context = self._get_context()
    with context.store().query().scoped() as q:
      for d in analyze_url_in(q.file_enum('lib/%', neg=True)):
        tentative = False
        if '...' in d['v']:
          ui.warn('truncated value found; disassemble again with wider fields', onetime=True)
          tentative = True
        self._helper.raise_issue(self._helper.build_issue(
          sigid='nat-detect-urls',
          cvss=self._cvss_info,
          title='detected {}'.format(d['typ']),
          cfd='tentative' if tentative else 'firm',
          info0=d['v'],
          aff0=d['fn'],
        ))

  @classmethod
  def _entropy_of(cls, string: str) -> float:
    o = 0.0
    m: Dict[str, int] = dict()
    for c in string:
      m[c] = m.get(c, 0) + 1
    for cnt in m.values():
      freq = float(cnt) / len(string)
      o -= freq * (math.log(freq) / math.log(2))
    return o

  @classmethod
  def _assumed_randomness_of(cls, string: str) -> float:
    try:
      return cls._entropy_of(string) / float(math.log(len(string)) / math.log(2))
    except ValueError:
      return 0



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.env import is_in_container
from trueseeing.core.ui import ui, FileTransferProgressReporter

if TYPE_CHECKING:
  from typing import Any, Optional, Literal, Dict
  from trueseeing.api import CommandHelper, Command, CommandMap, OptionMap

  ArchiveFormat = Optional[Literal['tar:', 'tar:gz']]

class AssembleCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return AssembleCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'ca':dict(e=self._assemble, n='ca[!] /path', d='assemble as target from path', t={'apk'}),
      'ca!':dict(e=self._assemble, t={'apk'}),
      'cd':dict(e=self._disassemble, n='cd[s][!] /path', d='disassemble target into path', t={'apk'}),
      'cd!':dict(e=self._disassemble, t={'apk'}),
      'cds':dict(e=self._disassemble_nodex, t={'apk'}),
      'cds!':dict(e=self._disassemble_nodex, t={'apk'}),
      'cm':dict(e=self._merge, n='cm[!] [output.apk]', d='merge slices into a plain apk', t={'xapk'}),
      'cm!':dict(e=self._merge, t={'xapk'}),
      'co':dict(e=self._export_context, n='co[!] /path [pat]', d='export codebase', t={'apk'}),
      'co!':dict(e=self._export_context, t={'apk'}),
    }

  def get_options(self) -> OptionMap:
    return {
      'nocache':dict(n='nocache', d='do not replicate content before build [ca]', t={'apk'}),
    }

  async def _assemble(self, args: deque[str]) -> None:
    apk = self._helper.require_target('need target (i.e. output apk filename)')

    cmd = args.popleft()

    if not args:
      ui.fatal('need root path')

    import os
    import re
    import time
    from tempfile import TemporaryDirectory
    from trueseeing.core.android.asm import APKAssembler
    from trueseeing.core.android.tools import move_apk

    root = args.popleft()
    origapk = re.sub(r'(\.x?apk)$', r'\1.orig', apk)

    stem = re.sub(r'\.x?apk$', '', apk)
    for typ in ['apk', 'xapk']:
      print(origapk, f'{stem}.{typ}.orig')
      if os.path.exists(f'{stem}.{typ}.orig') and not cmd.endswith('!'):
        ui.fatal('backup file exists; force (!) to overwrite')

    if apk.endswith('.xapk'):
      ui.warn('assembling xapk is not supported; assembling as merged apk')
      apk = apk.replace('.xapk', '.apk')

    opts = self._helper.get_effective_options(self._helper.get_modifiers(args))

    ui.info('assembling {root} -> {apk}'.format(root=root, apk=apk))

    at = time.time()

    with TemporaryDirectory() as td:
      archive = self._deduce_archive_format(root)

      if not archive and opts.get('nocache', not is_in_container()):
        self._warn_if_container('nocache could be slow in container builds (try assembling from archives)')
        path = root
      else:
        path = os.path.join(td, 'f')

        if not archive:
          self._warn_if_container('caching massive files could be slow in container builds (try assembling from archives)')

        with FileTransferProgressReporter('caching content').scoped() as progress:
          if not archive:
            from trueseeing.core.tools import copytree
            for nr in copytree(os.path.join(root, '.'), path, divisor=(256 if progress.using_bar() else 1024)):
              progress.update(nr)
            progress.done()
          elif archive.startswith('tar'):
            from trueseeing.core.tools import copy_from_pack
            prefix = 'files'
            for nr in copy_from_pack(root, td, prefix=prefix, divisor=(256 if progress.using_bar() else 1024)):
              progress.update(nr)
            os.rename(os.path.join(td, prefix), path)
            progress.done()

      outapk, outsig = await APKAssembler.assemble_from_path(td, path)

      if os.path.exists(apk):
        move_apk(apk, origapk)

      move_apk(outapk, apk)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  def _warn_if_container(self, *args: Any, **kw: Any) -> None:
    from trueseeing.core.env import is_in_container
    if is_in_container():
      ui.warn(*args, **kw)

  async def _disassemble(self, args: deque[str], nodex: bool = False) -> None:
    apk = self._helper.require_target()

    cmd = args.popleft()

    if not args:
      ui.fatal('need output path')

    import os
    import time
    from shutil import rmtree
    from tempfile import TemporaryDirectory
    from trueseeing.core.android.asm import APKDisassembler
    from trueseeing.core.tools import move_as_output, pack_as_output

    path = args.popleft()

    if os.path.exists(path):
      if not cmd.endswith('!'):
        ui.fatal('output path exists; force (!) to overwrite')
      else:
        try:
          rmtree(path)
        except NotADirectoryError:
          os.remove(path)

    archive = self._deduce_archive_format(path)

    if not archive:
      self._warn_if_container('disassembling to directory could be slow in container builds (try disassembling to archives)')

    ui.info('disassembling {apk} -> {path}{nodex}'.format(apk=apk, path=path, nodex=' [res]' if nodex else ''))

    at = time.time()

    with TemporaryDirectory() as td:
      await APKDisassembler.disassemble_to_path(apk, td, nodex=nodex, merge=apk.endswith('.xapk'))

      if not archive:
        with FileTransferProgressReporter('disassemble: writing').scoped() as progress:
          for nr in move_as_output(td, path, allow_orphans=True):
            progress.update(nr)
          progress.done()
      elif archive.startswith('tar'):
        with FileTransferProgressReporter('disassemble: writing').scoped() as progress:
          for nr in pack_as_output(td, path, prefix='files', subformat=archive[4:], allow_orphans=True):
            progress.update(nr)
          progress.done()
        pass

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))

  async def _disassemble_nodex(self, args: deque[str]) -> None:
    await self._disassemble(args, nodex=True)

  async def _export_context(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need path')

    root = args.popleft()
    ui.info('exporting target to {root}'.format(root=root))

    if args:
      pat = args.popleft()
    else:
      pat = None

    import os
    import time

    archive = self._deduce_archive_format(root)

    if not archive:
      self._warn_if_container('exporting to directory could be slow in container builds (try exporting to archives)')

    at = time.time()
    extracted = 0
    context = self._helper.get_context()
    q = context.store().query()

    if not archive:
      for path,blob in q.file_enum(pat=pat, regex=True):
        target = os.path.join(root, *path.split('/'))
        if extracted % 10000 == 0:
          ui.info(' .. {nr} files'.format(nr=extracted))
        os.makedirs(os.path.dirname(target), exist_ok=True)
        with open(target, 'wb') as f:
          f.write(blob)
          extracted += 1
    elif archive.startswith('tar'):
      import tarfile
      from io import BytesIO
      kwargs: Dict[str, int] = dict()
      subformat = archive[4:]

      if subformat in ['gz']:
        kwargs.update(dict(compresslevel=3))

      with tarfile.open(root, 'w:{}'.format(subformat), **kwargs) as tf:  # type: ignore[call-overload]
        now = int(time.time())
        for path,blob in q.file_enum(pat=pat, regex=True):
          target = os.path.join('files', *path.split('/'))
          if extracted % 10000 == 0:
            ui.info(' .. {nr} files'.format(nr=extracted))

          bf = BytesIO(blob)
          ti = tarfile.TarInfo(name=target)
          ti.size = len(blob)
          ti.uname = 'root'
          ti.gname = 'root'
          ti.mode = 0o600
          ti.mtime = now
          tf.addfile(ti, fileobj=bf)
          extracted += 1

    ui.success('done: {nr} files ({t:.02f} sec.)'.format(nr=extracted, t=(time.time() - at)))

  def _deduce_archive_format(self, path: str) -> ArchiveFormat:
    if path.endswith('.tar'):
      return 'tar:'
    elif path.endswith('.tar.gz'):
      return 'tar:gz'
    else:
      return None

  async def _merge(self, args: deque[str]) -> None:
    target = self._helper.require_target('need target')

    cmd = args.popleft()

    import os
    import time
    from tempfile import TemporaryDirectory
    from trueseeing.core.android.asm import APKAssembler
    from trueseeing.core.android.tools import move_apk

    if args:
      outfn = args.popleft()
    else:
      outfn = target.replace('.xapk', '.apk')

    origfn = outfn.replace('.apk', '.apk.orig')

    if os.path.exists(origfn) and not cmd.endswith('!'):
      ui.fatal('backup file exists; force (!) to overwrite')

    ui.info('merging slices {target} -> {outfn}'.format(target=target, outfn=outfn))

    at = time.time()

    with TemporaryDirectory() as td:
      apk, sig = await APKAssembler.merge_slices(target, td)

      if os.path.exists(outfn):
        move_apk(outfn, origfn)

      move_apk(apk, outfn)

    ui.success('done ({t:.02f} sec.)'.format(t=(time.time() - at)))



from __future__ import annotations
from typing import TYPE_CHECKING, overload

import os.path
from abc import ABC, abstractmethod
from functools import cache

if TYPE_CHECKING:
  from typing import List, Dict, Optional, Set, Literal, Any, Mapping, AsyncIterator, Tuple, Iterator
  from typing_extensions import Self
  from trueseeing.api import FormatEntry
  from trueseeing.core.store import Store
  from trueseeing.core.android.context import APKContext

  ContextType = str
  ContextInfo = Mapping[str, Any]

class FileOpener:
  _formats: Dict[str, FormatEntry]

  def __init__(self, force_opener: Optional[str] = None) -> None:
    from trueseeing.core.config import Configs
    self._confbag = Configs.get().bag
    self._formats = dict()
    self._force_opener = force_opener
    self._init_formats()

  def get_context(self, path: str) -> Context:
    from trueseeing.core.exc import InvalidFileFormatError
    if not self._force_opener:
      import re

      def _key(x: Tuple[str, FormatEntry]) -> int:
        try:
          return len(x[1]['r'])
        except KeyError:
          from trueseeing.core.ui import ui
          ui.warn('{} ({}): old-style format entry detected; update definition'.format(x[1]['d'], x[0]))
          return len(x[0])

      for k,v in sorted(self._formats.items(), key=_key, reverse=True):
        if not re.search(v.get('r', k), path, re.IGNORECASE):
          continue
        c = v['e'](path)
        if c is None:
          continue
        return c
      raise InvalidFileFormatError()
    else:
      if self._force_opener in self._formats:
        c = self._formats[self._force_opener]['e'](path)
        if c is not None:
          return c
      raise InvalidFileFormatError()

  def get_formats(self) -> Iterator[Mapping[str, str]]:
    for k,v in self._formats.items():
      yield dict(n=k, d=v['d'])

  def _init_formats(self) -> None:
    from trueseeing.core.ext import Extension

    self._formats.update({
      'apk':dict(e=self._handle_apk, r=r'\.apk$', d='Android application package'),
      'xapk':dict(e=self._handle_xapk, r=r'\.xapk$', d='Android appllication bundle'),
      'ipa': dict(e=self._handle_ipa, r=r'\.ipa$', d='iOS application archive'),
    })

    for clazz in Extension.get().get_fileformathandlers():
      t = clazz.create()
      self._formats.update(t.get_formats())
      self._confbag.update(t.get_configs())

  def _handle_apk(self, path: str) -> Optional[Context]:
    from trueseeing.core.android.context import APKContext
    return APKContext(path)

  def _handle_xapk(self, path: str) -> Optional[Context]:
    from trueseeing.core.android.context import XAPKContext
    return XAPKContext(path)

  def _handle_ipa(self, path: str) -> Optional[Context]:
    from trueseeing.core.ios.context import IPAContext
    return IPAContext(path)

class Context(ABC):
  _path: str
  _excludes: List[str]
  _store: Optional[Store] = None
  _wd: Optional[str] = None

  def __init__(self, path: str) -> None:
    self._path = os.path.realpath(path)
    self._excludes = []
    # FIXME: is there a strict need of caching here?
    _ = self.wd

  @property
  def target(self) -> str:
    return self._path

  @property
  def wd(self) -> str:
    if self._wd is None:
      self._wd = self._get_workdir()
    return self._wd

  @property
  def excludes(self) -> List[str]:
    return self._excludes

  @excludes.setter
  def excludes(self, v: List[str]) -> None:
    self._excludes = v

  @property
  def type(self) -> Set[ContextType]:
    return self._get_type()

  # XXX
  @overload
  def require_type(self, typ: Literal['apk']) -> APKContext: ...
  @overload
  def require_type(self, typ: ContextType) -> Context: ...
  def require_type(self, typ: ContextType) -> Any:
    if typ in self.type:
      return self
    from trueseeing.core.exc import InvalidContextError
    raise InvalidContextError()

  def store(self) -> Store:
    if self._store is None:
      assert self.wd is not None
      from trueseeing.core.store import Store
      self._store = Store(self.wd)
    return self._store

  def size_of(self) -> Optional[int]:
    return self._get_size()

  def fingerprint_of(self) -> str:
    return self._get_fingerprint()

  def remove(self) -> None:
    if os.path.exists(self.wd):
      from shutil import rmtree
      rmtree(self.wd)
    self._store = None

  def exists(self) -> bool:
    return os.path.isdir(self.wd)

  def create(self, exist_ok: bool = False) -> None:
    os.makedirs(self.wd, mode=0o700, exist_ok=exist_ok)

  def has_patches(self) -> bool:
    if self.exists():
      return self.store().query().patch_exists(None)
    else:
      return False

  def _get_identity(self) -> bytes:
    from os import stat
    c = stat(self._path)
    return f'0,{c.st_size},{c.st_mtime_ns}'.encode()

  def _get_analysis_flag_name(self, level: int) -> str:
    return f'.done{level}' if level < 4 else '.done'

  def get_analysis_level(self) -> int:
    for level in range(4, 0, -1):
      fn = os.path.join(self.wd, self._get_analysis_flag_name(level))
      if os.path.exists(fn):
        with open(fn, 'rb') as f:
          expected = f.read()
        if not expected:
          with open(fn, 'wb') as g:
            from trueseeing.core.ui import ui
            ui.warn('old style flags detected; marking them as for current target')
            g.write(self._get_identity())
          return level
        else:
          if expected == self._get_identity():
            return level
          else:
            return 0
    return 0

  def _mark_analysis_done(self, level: int) -> None:
    flagfn = self._get_analysis_flag_name(level)
    with open(os.path.join(self.wd, flagfn), 'wb') as f:
      f.write(self._get_identity())

  async def analyze(self, level: int = 4) -> Self:
    from trueseeing.core.ui import ui
    if self.get_analysis_level() >= level:
      await self._recheck_schema()
      ui.debug('analyzed once')
    else:
      if os.path.exists(self.wd):
        ui.info('analyze: removing leftover')
        self.remove()

      if level > 0:
        self.create()
        await self._analyze(level=level)
      else:
        self.invalidate()

      self._mark_analysis_done(level)
    return self

  def invalidate(self) -> None:
    for level in range(4, 0, -1):
      flag = os.path.join(self.wd, self._get_analysis_flag_name(level))
      if os.path.exists(flag):
        os.remove(flag)

  @abstractmethod
  def _get_workdir(self) -> str: ...

  @abstractmethod
  def _get_size(self) -> Optional[int]: ...

  @abstractmethod
  def _get_fingerprint(self) -> str: ...

  @abstractmethod
  def _get_type(self) -> Set[ContextType]: ...

  @abstractmethod
  async def _recheck_schema(self) -> None: ...

  @abstractmethod
  async def _analyze(self, level: int) -> None: ...

  async def _get_info(self, extended: bool) -> AsyncIterator[ContextInfo]:
    yield dict(path=self._path)
    size = self.size_of()
    if size:
      yield dict(size=size)
    yield dict(fp=self.fingerprint_of())
    yield dict(ctx=self.wd)
    yield dict(_patch=self.has_patches())
    yield dict(_analysis=self.get_analysis_level())

class Fingerprint:
  @cache
  def get(self, path: str) -> str:
    from hashlib import sha256
    with open(path, 'rb') as f:
      return sha256(f.read()).hexdigest()



from __future__ import annotations
from typing import TYPE_CHECKING
import os
import re

if TYPE_CHECKING:
    from typing import Iterable, Any, Iterator, Mapping, Tuple

def analyze_api(path: str) -> Iterator[Mapping[str, Any]]:
    def walker(path: str) -> Iterator[Tuple[str, bytes]]:
        for dirpath, _, filenames in os.walk(path):
            for fn in filenames:
                target = os.path.join(dirpath, fn)
                with open(target, 'rb') as f:
                    yield target, f.read()

    return analyze_api_in(walker(path))

def analyze_api_in(gen: Iterable[Tuple[str, bytes]]) -> Iterator[Mapping[str, Any]]:
    # XXX: focusing on oneline
    pats = rb'^([_.].*?:[0-9a-f]{8}) +?[0-9a-f]+ +?b[a-z]* +(.*?);(\[[A-Za-z]+ .*?\]|undefined _.*)$'
    blacklist = '|'.join([
        ' _objc_',
        ' _swift_',
        ' _OUTLINED_FUNCTION_[0-9]+',
        ' __cxa_',
        ' __swift_',
        ' __release_weak',
        ' ___cxa_',
        ' ___swift_',
        ' ___stack_chk_fail',
        ' ___chkstk_darwin',
    ])

    for fn, s in gen:
      for m in re.finditer(pats, s, flags=re.MULTILINE):
        origin = m.group(1).decode('latin1').strip('"').replace(':','+')
        target = m.group(2).decode('latin1').strip('"')
        call = m.group(3).decode('latin1').strip('"')
        if not re.search(blacklist, call):
          if call.startswith('['):
            lang = 'objc'
          elif call.startswith('undefined _$'):
            lang = 'swift'
          elif call.startswith('undefined __Z'):
            lang = 'cpp'
          else:
            lang = 'c'

          if 'EXTERNAL' in target:
            yield dict(fn=fn, origin=origin, typ='API', lang=lang, call=call)
          else:
            yield dict(fn=fn, origin=origin, typ='private', lang=lang, call=call)

def analyze_lib_needs_in(gen: Iterable[Tuple[str, bytes]]) -> Iterator[Mapping[str, Any]]:
    pats = rb"(@rpath)?/[0-9A-za-z/.]+(\.framework/[0-9A-za-z/.]+|.dylib)"

    seen = set()
    for n, s in gen:
      for m in re.finditer(pats, s[:1048576]):
          fnlike = m.group(0).decode('latin1').strip('"')
          if fnlike not in seen:
              seen.add(fnlike)
              yield dict(fn=n, v=fnlike)

def get_origin(n: str, l: bytes) -> Mapping[str, Any]:
    pat = rb'(_.*?:[0-9a-f]{8}(?: |[0-9a-f]))[0-9a-f]+? +[a-z]+ '
    m = re.match(pat, l)
    if m:
        origin = m.group(1).decode('latin1').strip('"')
        sect, offs = origin.split(':')
        return dict(fn=n, sect=sect, offs=int(offs, 16))
    else:
        raise ValueError()



from __future__ import annotations
from typing import TYPE_CHECKING
import os
import re

if TYPE_CHECKING:
    from typing import Iterable, Any, Iterator, Mapping, Tuple

def analyze_api(path: str) -> Iterator[Mapping[str, Any]]:
    def walker(path: str) -> Iterator[Tuple[str, bytes]]:
        for dirpath, _, filenames in os.walk(path):
            for fn in filenames:
                target = os.path.join(dirpath, fn)
                with open(target, 'rb') as f:
                    yield target, f.read()

    return analyze_api_in(walker(path))

def analyze_api_in(gen: Iterable[Tuple[str, bytes]]) -> Iterator[Mapping[str, Any]]:
    # XXX: focusing on oneline
    pats = rb'^([_.].*?:[0-9a-f]{8}) +?[0-9a-f]+ +?b[a-z]* +(.*?);([A-Za-z]+ .*)$'
    blacklist = '|'.join([
        r' (thunk_)?FUN_[0-9a-f]+',
        r' __cxa_',
        r' __stack_chk_fail',
        r' operator\.',
        r' ~',
    ])

    for fn, s in gen:
      for m in re.finditer(pats, s, flags=re.MULTILINE):
        origin = m.group(1).decode('latin1').strip('" ').replace(':','+')
        target = m.group(2).decode('latin1').strip('" ')
        call = m.group(3).decode('latin1').strip('" ')
        if not re.search(blacklist, call):
          if re.search(r'operator|::.*?::', call):
            lang = 'cpp'
          else:
            lang = 'c'

          if 'EXTERNAL' in target:
            yield dict(fn=fn, origin=origin, typ='API', lang=lang, call=call)
          else:
            yield dict(fn=fn, origin=origin, typ='private', lang=lang, call=call)



from __future__ import annotations
from typing import TYPE_CHECKING

if TYPE_CHECKING:
  from typing import ClassVar, Dict, Optional
  from trueseeing.api import ConfigEntry

class Configs:
  _i: ClassVar[Optional[Configs]] = None
  _bag: Dict[str, ConfigEntry]

  def __init__(self) -> None:
    self._bag = dict()

  @classmethod
  def get(cls) -> Configs:
    if not cls._i:
      cls._i = Configs()
    return cls._i

  @property
  def bag(self) -> Dict[str, ConfigEntry]:
    return self._bag



from __future__ import annotations
from typing import TYPE_CHECKING
from functools import cache

if TYPE_CHECKING:
  from typing import Type, Iterator
  from trueseeing.api import Command

@cache
def discover() -> Iterator[Type[Command]]:
  from trueseeing.api import Command
  from importlib import import_module
  from trueseeing.core.tools import get_public_subclasses, get_missing_methods, discover_modules_under

  for mod in discover_modules_under('trueseeing.app.cmd'):
    m = import_module(mod)
    for c in get_public_subclasses(m, Command, 'CommandMixin'):  # type:ignore[type-abstract]
      assert not get_missing_methods(c)
      yield c



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.core.ui import CoreProgressReporter, ScanProgressReporter, ui
from trueseeing.core.context import FileOpener

if TYPE_CHECKING:
  from typing import List, Optional
  from trueseeing.core.context import Context
  from trueseeing.core.report import ReportGenerator, ReportFormat

class ScanMode:
  _outfile: Optional[str] = None

  def __init__(self, target: str, outform: ReportFormat, outfile: Optional[str], sigsels: List[str] = [], excludes: List[str] = [], force_opener: Optional[str] = None) -> None:
    from trueseeing.core.scan import Scanner
    self._target = target
    self._outfile = outfile
    self._context = FileOpener(force_opener=force_opener).get_context(self._target)
    self._context.excludes = excludes
    self._reporter = self._get_reporter(self._context, outform, outfile)
    self._scanner = Scanner(self._context, sigsels=sigsels, excludes=excludes)

  async def reanalyze(self) -> int:
    with CoreProgressReporter().scoped():
      self._context.remove()
      await self._context.analyze()
      return 0

  async def scan(self, incremental: bool = False, oneshot: bool = False) -> int:
    from trueseeing.core.exc import InvalidSchemaError
    with CoreProgressReporter().scoped():
      import time
      at = time.time()
      try:
        try:
          await self._context.analyze()
        except InvalidSchemaError:
          ui.warn('invalid schema detected; forcing reanalysis')
          self._context.remove()
          await self._context.analyze()

        ui.info(f"{self._target} -> {self._context.wd}")

        with self._context.store().query().scoped() as q:
          if not incremental:
            await self._scanner.clear(q)

          with ScanProgressReporter().scoped():
            nr = await self._scanner.scan(q)

        if self._outfile is not None:
          if self._outfile == '-':
            from sys import stdout
            self._reporter.generate(stdout)
          else:
            with open(self._outfile, 'w') as f:
              self._reporter.generate(f)

        ui.success('{fn}: analysis done, {nr} issues ({t:.02f} sec.)'.format(fn=self._target, nr=nr, t=(time.time() - at)))

        return 1 if nr else 0
      finally:
        if oneshot:
          self._context.remove()

  @classmethod
  def _get_reporter(cls, context: Context, format: ReportFormat, fn: Optional[str]) -> ReportGenerator:
    from trueseeing.core.report import CIReportGenerator, JSONReportGenerator, HTMLReportGenerator
    if fn is None:
      return CIReportGenerator(context)
    else:
      if format == 'json':
        return JSONReportGenerator(context)
      else:
        return HTMLReportGenerator(context)



from __future__ import annotations
from typing import TYPE_CHECKING
from abc import ABC, abstractmethod

if TYPE_CHECKING:
  from collections import deque
  from typing import Any, TypedDict, Protocol, Optional, Callable, Coroutine, Union, List, Mapping, overload, Literal, Set
  from typing_extensions import deprecated
  from trueseeing.core.context import Context, ContextType
  from trueseeing.core.android.context import APKContext
  from trueseeing.core.model.issue import Issue, IssueConfidence

  ModifierEvent = Literal['begin', 'end']

  CommandEntrypoint = Callable[[deque[str]], Coroutine[Any, Any, None]]
  CommandlineEntrypoint = Callable[[str], Coroutine[Any, Any, None]]
  CommandPatternEntrypoints = Union[CommandEntrypoint, CommandlineEntrypoint]
  ModifierListenerEntrypoint = Callable[[ModifierEvent, str], Coroutine[Any, Any, None]]
  SignatureEntrypoint = Callable[[], Coroutine[Any, Any, None]]
  FormatHandlerEntrypoint = Callable[[str], Optional[Context]]
  ConfigGetterEntrypoint = Callable[[], Any]
  ConfigSetterEntrypoint = Callable[[Any], None]

  class Entry(TypedDict, total=False):
    e: CommandEntrypoint
    n: str
    d: str
    t: Set[str]

  class CommandEntry(Entry):
    pass

  class CommandPatternEntry(CommandEntry, total=False):
    e: CommandPatternEntrypoints  # type: ignore[misc]
    raw: bool

  class OptionEntry(Entry):
    pass

  class ModifierEntry(Entry):
    e: Optional[ModifierListenerEntrypoint]  # type: ignore[misc]

  class ConfigEntry(TypedDict):
    g: ConfigGetterEntrypoint
    s: ConfigSetterEntrypoint
    n: str
    d: str

  class SignatureEntry(TypedDict):
    e: SignatureEntrypoint
    d: str

  class FormatEntry(TypedDict):
    e: FormatHandlerEntrypoint
    r: str
    d: str

  CommandMap = Mapping[str, CommandEntry]
  CommandPatternMap = Mapping[str, CommandPatternEntry]
  OptionMap = Mapping[str, OptionEntry]
  ModifierMap = Mapping[str, ModifierEntry]
  ConfigMap = Mapping[str, ConfigEntry]
  SignatureMap = Mapping[str, SignatureEntry]
  FormatMap = Mapping[str, FormatEntry]

  class CommandHelper(Protocol):
    def get_target(self) -> Optional[str]: ...
    def require_target(self, msg: Optional[str] = None) -> str: ...
    def get_context_type(self) -> Optional[Set[ContextType]]: ...
    @overload
    def get_context(self) -> Context: ...
    @overload
    @deprecated('use get_context().require_type(...)')
    def get_context(self, typ: Literal['apk']) -> APKContext: ...
    @overload
    @deprecated('use get_context().require_type(...)')
    def get_context(self, typ: ContextType) -> Context: ...
    @overload
    @deprecated('use get_context().analyze(...)')
    async def get_context_analyzed(self, *, level: int = 4) -> Context: ...
    @overload
    @deprecated('use get_context().require_type(...).analyze(...)')
    async def get_context_analyzed(self, typ: Literal['apk'], *, level: int = 4) -> APKContext: ...
    @overload
    @deprecated('use get_context().require_type(...).analyze(...)')
    async def get_context_analyzed(self, typ: ContextType, *, level: int = 4) -> Context: ...
    def decode_analysis_level(self, level: int) -> str: ...
    async def run(self, s: str) -> None: ...
    async def run_cmd(self, tokens: deque[str], line: Optional[str]) -> bool: ...
    def get_modifiers(self, args: deque[str]) -> List[str]: ...
    def get_effective_options(self, mods: List[str]) -> Mapping[str, str]: ...
    def get_graph_size_limit(self, mods: List[str]) -> Optional[int]: ...
    def get_config(self, k: str) -> Any: ...
    def set_config(self, k: str, v: Any) -> None: ...

  class SignatureHelper(Protocol):
    @overload
    def get_context(self) -> Context: ...
    @overload
    @deprecated('use get_context().require_type(...)')
    def get_context(self, typ: Literal['apk']) -> APKContext: ...
    @overload
    @deprecated('use get_context().require_type(...)')
    def get_context(self, typ: ContextType) -> Context: ...
    def get_config(self, k: str) -> Any: ...
    def set_config(self, k: str, v: Any) -> None: ...
    def raise_issue(self, issue: Issue) -> None: ...
    def build_issue(
        self,
        sigid: str,
        cvss: str,
        title: str,
        cfd: IssueConfidence = 'firm',
        summary: Optional[str] = None,
        desc: Optional[str] = None,
        ref: Optional[str] = None,
        sol: Optional[str] = None,
        info0: Optional[str] = None,
        info1: Optional[str] = None,
        info2: Optional[str] = None,
        aff0: Optional[str] = None,
        aff1: Optional[str] = None,
        aff2: Optional[str] = None,
    ) -> Issue: ...


class FileFormatHandler(ABC):
  """File format handlers; they are responsible for opening path and create a valid Context."""

  @staticmethod
  @abstractmethod
  def create() -> FileFormatHandler:
    """Creates and return itself; This is because Python checks classes of concreteness only in their instantiation. e.g.

    return FooFileFormatHandler()
    """
    ...
  @abstractmethod
  def get_formats(self) -> FormatMap:
    """Creates and return format descriptor. Format descriptors are dicts comprise of:

    {"<path pattern>": dict(e=<entrypoint>, d="<description>")}

    e.g.:

    return {"\\.apk$": dict(e=self._handler, d="The Android Package File")}
    """
    ...
  @abstractmethod
  def get_configs(self) -> ConfigMap:
    """Creates and return the config descriptor it provides. Config descriptors are dicts comprise of:

    {"<name>": dict(g=<getter>, s=<setter>, n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      'some_value':dict(g=self._getter, s=self._setter, n='some_value=value', d='config variable'),
    }
    """
    ...

class Command(ABC):
  """Commands; they provides one or more interactive mode commands."""

  @staticmethod
  @abstractmethod
  def create(helper: CommandHelper) -> Command:
    """Creates and return itself; This is because Python checks classes of concreteness only in their instantiation. e.g.

    return FooCommand(helper)
    """
    ...
  @abstractmethod
  def get_commands(self) -> CommandMap:
    """Creates and return the command descriptor it provides. Command descriptors are dicts comprise of:

    {"<cmd>": dict(e=<entrypoint>, n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      "a": dict(e=self._analyze, n="a[a]", d="Analyze"),
      "aa": dict(e=self._analyze2),
    }
    """
    ...
  @abstractmethod
  def get_command_patterns(self) -> CommandPatternMap:
    """Creates and return the command pattern descriptor it provides. Command patterns are patterns that the first token (raw=False, the default) or whole command line (raw=True) should match. Entrypoints are called with tokens (raw=False) or whole command line (raw=True). Trailers are ignored while matching whole command line. Command pattern descriptors are dicts comprise of:

    {"<pattern>": dict(e=<entrypoint>, raw=<False|True>, n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      "\\$[a-z0-9]+": dict(e=self._alias, n="$<name>[=<value>]", d="define alias"),
      '\\(.+\\)':dict(e=self._alias2, raw=True, n='(macro x y; cmd; cmd; ..)', d='define macro'),
    }
    """
    ...
  @abstractmethod
  def get_modifiers(self) -> ModifierMap:
    """Creates and return the modifier descriptor it recognizes. Modifier descriptors are dicts comprise of:

    {"<name>": dict(n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      's':dict(n='@s:sig', d='include sig'),
      'x':dict(n='@x:pa.ckage.name', d='exclude package'),
    }
    """
    ...
  @abstractmethod
  def get_configs(self) -> ConfigMap:
    """Creates and return the config descriptor it provides. Config descriptors are dicts comprise of:

    {"<name>": dict(g=<getter>, s=<setter>, n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      'some_value':dict(g=self._getter, s=self._setter, n='some_value=value', d='config variable'),
    }
    """
    ...
  @abstractmethod
  def get_options(self) -> OptionMap:
    """Creates and return the option descriptor it recognizes. Option descriptors are dicts comprise of:

    {"<name>": dict(n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      'nocache':dict(n='nocache', d='do not replicate content before build [ca]')
    }
    """
    ...

class Signature(ABC):
  """Signatures; they provides one or more signatures for scanner."""

  @staticmethod
  @abstractmethod
  def create(helper: SignatureHelper) -> Signature:
    """Creates and return itself; This is because Python checks classes of concreteness only in their instantiation. e.g.

    return FooDetector(helper)
    """
    ...
  @abstractmethod
  def get_sigs(self) -> SignatureMap:
    """Creates and return signature descriptor. Signature descriptors are dicts comprise of:

    {"<signature id>": dict(e=<entrypoint>, d="<description>")}

    e.g.:

    return {"my-crypto-static-keys": dict(e=self._detect, d='Detects cryptographic function usage with static keys')}
    """
    ...
  @abstractmethod
  def get_configs(self) -> ConfigMap:
    """Creates and return the config descriptor it provides. Config descriptors are dicts comprise of:

    {"<name>": dict(g=<getter>, s=<setter>, n="<mnemonic>", d="<description>")}

    e.g.:

    return {
      'some_value':dict(g=self._getter, s=self._setter, n='some_value=value', d='config variable'),
    }
    """
    ...



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from trueseeing.api import Command, CommandHelper, CommandMap

class AnalyzeCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return AnalyzeCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'a':dict(e=self._analyze, n='a[a][a][!]', d='analyze target (aa: marginal, aaa: full)'),
      'a!':dict(e=self._analyze),
      'aa':dict(e=self._analyze2),
      'aa!':dict(e=self._analyze2),
      'aaa':dict(e=self._analyze3),
      'aaa!':dict(e=self._analyze3),
    }

  async def _analyze(self, args: deque[str], level: int = 2) -> None:
    target = self._helper.require_target()

    cmd = args.popleft()

    ui.info(f"analyzing {target}")

    context = self._helper.get_context()
    if cmd.endswith('!'):
      context.remove()
    await context.analyze(level=level)

  async def _analyze2(self, args: deque[str]) -> None:
    await self._analyze(args, level=3)

  async def _analyze3(self, args: deque[str]) -> None:
    await self._analyze(args, level=4)



from __future__ import annotations
from typing import TYPE_CHECKING

from contextlib import contextmanager

if TYPE_CHECKING:
  from pathlib import Path
  from typing import TypeVar, Iterator, TypedDict
  T = TypeVar('T')

  class Toolchain(TypedDict):
    apkeditor: Path
    apksigner: Path
    abe: Path

@contextmanager
def toolchains() -> Iterator[Toolchain]:
  from trueseeing.core.tools import require_in_path
  from importlib.resources import files, as_file
  require_in_path('java', 'java -version')
  libs = files('trueseeing')/'libs'/'android'
  with as_file(libs/'apkeditor.jar') as apkeditorpath:
    with as_file(libs/'apksigner.jar') as apksignerpath:
      with as_file(libs/'abe.jar') as abepath:
        yield dict(
          apkeditor=apkeditorpath,
          apksigner=apksignerpath,
          abe=abepath,
        )

def move_apk(src: str, dest: str) -> None:
  import shutil
  shutil.move(src, dest)
  try:
    shutil.move(src.replace('.apk', '.apk.idsig'), dest.replace('.apk', '.apk.idsig'))
  except OSError:
    pass



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.core.db import Query, StorePrep
from trueseeing.core.android.model import Op

if TYPE_CHECKING:
  from typing import Optional, Iterator
  from trueseeing.core.store import Store
  from trueseeing.core.android.model import InvocationPattern, Call

class APKStorePrep(StorePrep):
  def stage1(self) -> None:
    super().stage1()
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'android'/'store.0.sql').read_text())

  def stage2(self) -> None:
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'android'/'store.1.sql').read_text())

  def _get_cache_schema_id(self) -> int:
    return super()._get_cache_schema_id() ^ 0x48d42899

class APKQuery(Query):
  def __init__(self, store: Store) -> None:
    super().__init__(store)
    if store:
      from trueseeing.core.android.store import APKStore
      assert isinstance(store, APKStore)

  def _get_smali_forward_like_pattern_of_package(self, name: str, *, regex: bool = False) -> str:
    return r'L{}/{}'.format(name.replace('.', '/'), '' if regex else '%')

  def consts(self, pat: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join xref_const using (addr) where insn like :insn and sym regexp :pat', dict(insn=f'{pat.insn}%', pat=pat.value)):
      yield Op(addr, l)

  def consts_in_package(self, name: str, pat: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join xref_const using (addr) join map on (addr between low and high) where class like :pkg and insn like :insn and sym regexp :pat', dict(insn=f'{pat.insn}%', pat=pat.value, pkg=self._get_smali_forward_like_pattern_of_package(name))):
      yield Op(addr, l)

  def consts_in_class(self, addr: int, pat: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops join (select low, high from ops join map on (map_id=id) where addr=:addr) on (addr between low and high) join xref_const using (addr) where insn like :insn and sym regexp :pat", dict(addr=addr, insn=f'{pat.insn}%', pat=pat.value)):
      yield Op(addr, l)

  def sputs(self, target: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join xref_sput using (addr) where sym=:sym', dict(sym=target)):
      yield Op(addr, l)

  def iputs(self, target: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join xref_iput using (addr) where sym=:sym', dict(sym=target)):
      yield Op(addr, l)

  def op_get(self, addr: int) -> Op:
    for addr, l in self.db.execute("select addr, l from ops where addr=:addr", dict(addr=addr)):
      return Op(addr, l)
    raise IndexError()

  def ops_of(self, insn: str) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops join map on (map_id=id) where method is not null and l like :pat", dict(pat=f'    {insn} %')):
      yield Op(addr, l)

  def invocations(self, pat: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops join xref_invoke using (addr) where insn like :insn and sym regexp :pat", dict(insn=f'{pat.insn}%', pat=pat.value)):
      yield Op(addr, l)

  def invocations_in_class(self, addr: int, pat: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops join (select low, high from ops join map on (map_id=id) where addr=:addr) on (addr between low and high) join xref_invoke using (addr) where insn like :insn and sym regexp :pat", dict(addr=addr, insn=f'{pat.insn}%', pat=pat.value)):
      yield Op(addr, l)

  def callers_of(self, addr: int) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops join xref_invoke using (addr) where target=(select low from ops join map on (map_id=id) where method is not null and addr=:addr)", dict(addr=addr)):
      yield Op(addr, l)

  def qualname_of(self, addr: int) -> Optional[str]:
    for n, in self.db.execute("select class||'->'||method from ops join map on (map_id=id) where method is not null and addr=:addr", dict(addr=addr)):
      return n # type: ignore[no-any-return]
    return None

  def class_name_of(self, addr: int) -> Optional[str]:
    for n, in self.db.execute("select class from ops join map on (map_id=id) where addr=:addr", dict(addr=addr)):
      return n # type: ignore[no-any-return]
    return None

  def method_call_target_of(self, addr: Optional[int]) -> Optional[str]:
    if addr is not None:
      for sym, in self.db.execute('select sym from xref_invoke where addr=:addr', dict(addr=addr)):
        return sym # type: ignore[no-any-return]
    return None

  def reversed_insns_in_method(self, addr: int) -> Iterator[Op]:
    for addr, l in self.db.execute("select addr, l from ops where addr between (select low from ops join map on (map_id=id) where addr=:addr) and :addr-1 order by addr desc", dict(addr=addr)):
      yield Op(addr, l)

  def in_same_mod(self, addr0: int, addr1: int) -> bool:
    for _, in self.db.execute("select 1 from ops where addr=:addr0 and map_id=(select map_id from ops where addr=:addr1)", dict(addr0=addr0, addr1=addr1)):
      return True
    return False

  def methods_in_class(self, method_name: str, related_class_name: str) -> Iterator[int]:
    for addr, in self.db.execute('select low from map where method like :method_pat and class in (select class from class_rel where super like :class_pat or impl like :class_pat)', dict(class_pat=f'%{related_class_name}%', method_pat=f'%{method_name}%')):
      yield addr

  def related_classes(self, related_class_name: str) -> Iterator[int]:
    for addr, in self.db.execute('select low from map where class in (select class from class_rel where super regexp :pat or impl regexp :pat)', dict(pat=related_class_name)):
      yield addr

  def matches_in_method(self, addr: int, pattern: InvocationPattern) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join (select low, high from ops join map on (map_id=id) where addr=:addr) on (addr between low and high) where l regexp :pat', dict(addr=addr, pat=f'^  ${pattern.insn} ${pattern.value}')):
      yield Op(addr, l)

  def methods_with_modifier(self, pattern: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join map on (addr=low) where method is not null and l like :pat', dict(pat=f'% {pattern} %')):
      yield Op(addr, l)

  def classes_has_method_named(self, pattern: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join map on (addr=low) where method is null and class in (select class from map where method regexp :pat)', dict(pat=pattern)):
      yield Op(addr, l)

  def classes_extends_has_method_named(self, method: str, extends: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join map on (addr=low) where method is null and class in (select class from map join class_rel using (class) where method regexp :method and super regexp :extends)', dict(method=method, extends=extends)):
      yield Op(addr, l)

  def classes_implements_has_method_named(self, method: str, implements: str) -> Iterator[Op]:
    for addr, l in self.db.execute('select addr, l from ops join map on (addr=low) where method is null and class in (select class from map join class_rel using (class) where method regexp :method and impl regexp :implements)', dict(method=method, implements=implements)):
      yield Op(addr, l)

  def class_names(self, pat: str) -> Iterator[str]:
    for name, in self.db.execute('select class from map where method is null and class regexp :pat', dict(pat=pat)):
      yield name

  def body(self, class_name: str, method_name: Optional[str]) -> Iterator[Op]:
    stmt0 = 'select addr, l from ops join map on (addr between low and high) where method is null and class=:class_name'
    stmt1 = 'select addr, l from ops join map on (addr between low and high) where method=:method_name and class=:class_name'
    for addr, l in self.db.execute(stmt1 if method_name else stmt0, dict(class_name=class_name, method_name=method_name)):
      yield Op(addr, l)

  def call_add_batch(self, gen: Iterator[Call]) -> None:
    stmt0 = 'insert into ncalls (priv, cpp, target, path, sect, offs) values (:priv, :cpp, :target, :path, :sect, :offs)'
    self.db.executemany(stmt0, gen)

  def call_count(self) -> int:
    stmt0 = 'select count(1) from ncalls'
    for n, in self.db.execute(stmt0):
      return n # type:ignore[no-any-return]
    return 0

  def calls(self, priv: bool = False, api: bool = False) -> Iterator[Call]:
    stmt0 = 'select priv, cpp, target, path, sect, offs from ncalls'
    stmt1 = 'select priv, cpp, target, path, sect, offs from ncalls where priv=:is_priv'
    for priv, cpp, target, path, sect, offs in self.db.execute(stmt1 if (priv or api) else stmt0, dict(is_priv=priv)):
      yield dict(
        path=path,
        sect=sect,
        offs=offs,
        priv=priv,
        cpp=cpp,
        target=target,
      )



from __future__ import annotations
from typing import TYPE_CHECKING

import re
from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui, OpFormatter, OpLister

if TYPE_CHECKING:
  from trueseeing.api import CommandHelper, Command, CommandMap

class SearchCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return SearchCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '/c':dict(e=self._search_call, n='/c [pat]', d='search call for pattern', t={'apk'}),
      '/k':dict(e=self._search_const, n='/k insn [pat]', d='search consts for pattern', t={'apk'}),
      '/p':dict(e=self._search_put, n='/p[i] [pat]', d='search s/iputs for pattern', t={'apk'}),
      '/dp':dict(e=self._search_defined_package, n='/dp [pat]', d='search packages matching pattern', t={'apk'}),
      '/dc':dict(e=self._search_defined_class, n='/dc [pat]', d='search classes matching pattern', t={'apk'}),
      '/dcx':dict(e=self._search_derived_class, n='/dcx classpat [methpat]', d='search classes defining methods and extending ones matching patterns', t={'apk'}),
      '/dci':dict(e=self._search_implementing_class, n='/dci ifacepat [methpat]', d='search classes defining methods and implementing interfaces matching patterns', t={'apk'}),
      '/dm':dict(e=self._search_defined_method, n='/dm pat', d='search classes defining methods matching pattern', t={'apk'}),
    }

  async def _search_call(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need pattern')

    pat = args.popleft()

    from trueseeing.core.android.model import InvocationPattern
    context = await self._helper.get_context().require_type('apk').analyze()

    ui.info(f'searching call: {pat}')

    with context.store().query().scoped() as q:
      OpLister(OpFormatter(q)).list_tagged(q.invocations(InvocationPattern('invoke-', pat)))

  async def _search_const(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need insn')

    insn = args.popleft()
    if args:
      pat = args.popleft()
    else:
      pat = '.'

    from trueseeing.core.android.model import InvocationPattern
    context = await self._helper.get_context().require_type('apk').analyze()

    ui.info(f'searching const: {pat} [{insn}]')

    with context.store().query().scoped() as q:
      OpLister(OpFormatter(q)).list_tagged(q.consts(InvocationPattern(insn, pat)))

  async def _search_put(self, args: deque[str]) -> None:
    self._helper.require_target()

    cmd = args.popleft()

    if args:
      pat = args.popleft()
    else:
      pat = '.'

    context = await self._helper.get_context().require_type('apk').analyze()
    q = context.store().query()
    if not cmd.endswith('i'):
      fun = q.sputs
      funtype = 's'
    else:
      fun = q.iputs
      funtype = 'i'

    ui.info(f'searching {funtype}puts: {pat}')

    OpLister(OpFormatter(q)).list_tagged(fun(pat))

  async def _search_defined_package(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if args:
      pat = args.popleft()
    else:
      pat = '.'

    import os
    context = await self._helper.get_context().require_type('apk').analyze()

    ui.info(f'searching packages: {pat}')

    packages = set()
    for fn in (context.source_name_of_disassembled_class(r) for r in context.disassembled_classes()):
      if fn.endswith('.smali'):
        packages.add(os.path.dirname(fn))
    for pkg in sorted(packages):
      if re.match(pat, pkg):
        ui.info(pkg)

  async def _search_defined_class(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if args:
      pat = args.popleft()
    else:
      pat = '.'

    context = await self._helper.get_context().require_type('apk').analyze()

    ui.info(f'searching classes: {pat}')

    with context.store().query().scoped() as q:
      for name in q.class_names(pat):
        ui.info(name)

  async def _search_derived_class(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need class')

    base = args.popleft()
    if args:
      pat = args.popleft()
    else:
      pat = '.'

    context = await self._helper.get_context().require_type('apk').analyze()

    if pat == '.':
      ui.info(f'searching classes deriving from: {base}')
    else:
      ui.info(f'searching classes deriving from: {base} [has method:{pat}]')

    with context.store().query().scoped() as q:
      OpLister(OpFormatter(q)).list_untagged(q.classes_extends_has_method_named(pat, base))

  async def _search_implementing_class(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need pattern')

    interface = args.popleft()
    if args:
      pat = args.popleft()
    else:
      pat = '.'

    if pat == '.':
      ui.info(f'searching classes implementing: {interface}')
    else:
      ui.info(f'searching classes implementing: {interface} [has method:{pat}]')

    context = await self._helper.get_context().require_type('apk').analyze()
    q = context.store().query()
    OpLister(OpFormatter(q)).list_untagged(q.classes_implements_has_method_named(pat, interface))

  async def _search_defined_method(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if not args:
      ui.fatal('need classname')

    pat = args.popleft()

    ui.info(f'searching classes defining method: {pat}')

    context = await self._helper.get_context().require_type('apk').analyze()
    q = context.store().query()
    OpLister(OpFormatter(q)).list_untagged(q.classes_has_method_named(pat))



from __future__ import annotations
from typing import TYPE_CHECKING

from functools import cache
from trueseeing.core.env import get_adb_host
from trueseeing.core.tools import invoke, invoke_passthru, invoke_streaming
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Optional, AsyncIterator, Any

class AndroidDevice:
  def __init__(self) -> None:
    pass

  async def invoke_adb(self, cmd: str, **kwargs: Any) -> str:
    self._require_adb()
    line = self._get_adb_cmdline(cmd)
    ui.debug("invoking: {line}")
    return await invoke(line, **kwargs)

  async def invoke_adb_passthru(self, cmd: str, **kwargs: Any) -> None:
    self._require_adb()
    line = self._get_adb_cmdline(cmd)
    ui.debug("invoking: {line}")
    await invoke_passthru(line, **kwargs)

  async def invoke_adb_streaming(self, cmd: str, **kwargs: Any) -> AsyncIterator[bytes]:
    self._require_adb()
    line = self._get_adb_cmdline(cmd)
    ui.debug("invoking: {line}")
    async for l in invoke_streaming(line, **kwargs):
      yield l

  def get_adb_cmdline(self, cmd: str) -> str:
    return self._get_adb_cmdline(cmd)

  def _get_adb_cmdline(self, cmd: str) -> str:
    host: Optional[str] = get_adb_host()

    line = 'adb {host} {cmd}'.format(
      host=f'-L {host}' if host else '',
      cmd=cmd
    )
    return line

  def require_adb(self) -> None:
    return self._require_adb()

  @cache
  def _require_adb(self) -> None:
    from trueseeing.core.tools import require_in_path
    require_in_path('adb', 'adb version')

  async def is_fullbackup_available(self) -> bool:
    out = await self.invoke_adb('backup 2>&1 || exit 0')
    return 'backup either' in out

  async def is_package_debuggable(self, package: str) -> bool:
    out = await self.invoke_adb(f"shell 'run-as {package} ls' 2>&1 || exit 0")
    return 'package not debuggable' not in out



from __future__ import annotations
from typing import TYPE_CHECKING

import re

if TYPE_CHECKING:
  from trueseeing.core.model.issue import IssueSeverity, IssueConfidence

class CVSS3Scoring:
  def __init__(self, m: re.Match[str]):
    self._m = m

  @classmethod
  def severity_of(cls, score: float) -> IssueSeverity:
    if score <= 0.0:
      return 'info'
    elif score < 4.0:
      return 'low'
    elif score < 7.0:
      return 'medium'
    elif score < 9.0:
      return 'high'
    else:
      return 'critical'

  @classmethod
  def temporalified(cls, vec: str, confidence: IssueConfidence) -> str:
    return '{v}RC:{c}/'.format(
      v=vec,
      c={'certain':'C','firm':'R','tentative':'U'}[confidence]
    )

  @classmethod
  def score_of(cls, vec: str) -> float:
    m = re.match(r'CVSS:3.0/AV:(?P<AV>[NALP])/AC:(?P<AC>[LH])/PR:(?P<PR>[NLH])/UI:(?P<UI>[NR])/S:(?P<S>[CU])/C:(?P<C>[HLN])/I:(?P<I>[HLN])/A:(?P<A>[HLN])(?:/RC:(?P<RC>[XCRU]))?/', vec)
    if m:
      return cls(m)._score()
    else:
      raise ValueError()

  def _score(self) -> float:
    return self._temporal_score()

  def _temporal_score(self) -> float:
    return self._roundup(self._base_score() * self._exploit_code_maturity_score() * self._remediation_level_score() * self._report_confidence_score())

  def _exploit_code_maturity_score(self) -> float:
    return 1

  def _remediation_level_score(self) -> float:
    return 1

  def _report_confidence_score(self) -> float:
    m = self._m.group('RC')
    if m:
      M = dict(X=1.0, C=1.0, R=0.96, U=0.92)
      return M[m]
    else:
      return 1

  def _base_score(self) -> float:
    impact, exploitability = self._subscore_impact(), self._subscore_exploitability()
    if impact <= 0:
      return 0
    else:
      if not self._scope_changed():
        return self._roundup(min(impact + exploitability, 10))
      else:
        return self._roundup(min(1.08 * (impact + exploitability), 10))

  def _subscore_impact(self) -> float:
    base = self._subscore_impact_base()
    if not self._scope_changed():
      return 6.42 * base
    else:
      return 7.52*(base-0.029) - 3.25*(base-0.02)**15

  def _subscore_impact_base(self) -> float:
    M = dict(N=0, L=0.22, H=0.56)
    C, I, A = M[self._m.group('C')], M[self._m.group('I')], M[self._m.group('A')]
    return 1 - ((1-C) * (1-I) * (1-A))

  def _subscore_exploitability(self) -> float:
    M_AV = dict(N=0.85, A=0.62, L=0.55, P=0.2)
    M_AC = dict(L=0.77, H=0.44)
    M_PR = dict(N=0.85, L=0.68 if self._scope_changed() else 0.62, H=0.50 if self._scope_changed() else 0.27)
    M_UI = dict(N=0.85, R=0.62)
    AV, AC, PR, UI = M_AV[self._m.group('AV')], M_AC[self._m.group('AC')], M_PR[self._m.group('PR')], M_UI[self._m.group('UI')]

    return 8.22 * AV * AC * PR * UI

  def _scope_changed(self) -> bool:
    return (self._m.group('S') == 'C')

  @classmethod
  def _roundup(cls, v: float) -> float:
    from math import ceil
    return ceil(v * 10.0) / 10.0



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Optional
  from trueseeing.api import CommandHelper, Command, CommandMap

class ReportCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ReportCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'gh':dict(e=self._report_html, n='gh[!] [report.html]', d='generate report (HTML)'),
      'gh!':dict(e=self._report_html),
      'gj':dict(e=self._report_json, n='gj[!] [report.json]', d='generate report (JSON)'),
      'gj!':dict(e=self._report_json),
      'gt':dict(e=self._report_text, n='gt[!] [report.txt]', d='generate report (text)'),
      'gt!':dict(e=self._report_text),
    }

  async def _report_html(self, args: deque[str]) -> None:
    outfn: Optional[str] = None

    self._helper.require_target()

    cmd = args.popleft()

    if args:
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    from trueseeing.core.report import HTMLReportGenerator
    context = self._helper.get_context()
    gen = HTMLReportGenerator(context)
    if outfn is None:
      from io import StringIO
      f0 = StringIO()
      gen.generate(f0)
      ui.stdout(f0.getvalue())
    else:
      with open(outfn, 'w') as f1:
        gen.generate(f1)

  async def _report_json(self, args: deque[str]) -> None:
    outfn: Optional[str] = None

    self._helper.require_target()

    cmd = args.popleft()

    if args:
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    from trueseeing.core.report import JSONReportGenerator
    context = self._helper.get_context()
    gen = JSONReportGenerator(context)
    if outfn is None:
      from io import StringIO
      f0 = StringIO()
      gen.generate(f0)
      ui.stdout(f0.getvalue())
    else:
      with open(outfn, 'w') as f1:
        gen.generate(f1)

  async def _report_text(self, args: deque[str]) -> None:
    outfn: Optional[str] = None

    self._helper.require_target()

    cmd = args.popleft()

    if args:
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    from trueseeing.core.report import CIReportGenerator
    context = self._helper.get_context()
    gen = CIReportGenerator(context)
    if outfn is None:
      from io import StringIO
      f0 = StringIO()
      gen.generate(f0)
      ui.stdout(f0.getvalue())
    else:
      with open(outfn, 'w') as f1:
        gen.generate(f1)



from __future__ import annotations
from typing import TYPE_CHECKING

import re

from trueseeing.core.android.model import InvocationPattern
from trueseeing.core.android.analyze.flow import DataFlow
from trueseeing.core.android.model import SignatureMixin

if TYPE_CHECKING:
  from typing import Optional
  from trueseeing.core.android.model import Op
  from trueseeing.core.android.db import APKQuery
  from trueseeing.api import Signature, SignatureHelper, SignatureMap

class PrivacyDeviceIdDetector(SignatureMixin):
  _id = 'privacy-device-id'
  description = 'Detects device fingerprinting behavior'
  _cvss = 'CVSS:3.0/AV:N/AC:H/PR:L/UI:R/S:C/C:L/I:N/A:N/'
  _summary = 'privacy concerns'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return PrivacyDeviceIdDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects device fingerprinting behavior')}

  def analyzed(self, q: APKQuery, op: Op) -> Optional[str]:
    x = self._an.get_param(op, 1).v
    if re.search(r'Landroid/provider/Settings\$Secure;->getString\(Landroid/content/ContentResolver;Ljava/lang/String;\)Ljava/lang/String;', x):
      try:
        if DataFlow(q).solved_constant_data_in_invocation(op, 1) == 'android_id':
          return 'ANDROID_ID'
        else:
          return None
      except DataFlow.NoSuchValueError:
        return None
    elif re.search(r'Landroid/telephony/TelephonyManager;->getDeviceId\(\)Ljava/lang/String;', x):
      return 'IMEI'
    elif re.search(r'Landroid/telephony/TelephonyManager;->getSubscriberId\(\)Ljava/lang/String;', x):
      return 'IMSI'
    elif re.search(r'Landroid/telephony/TelephonyManager;->getLine1Number\(\)Ljava/lang/String;', x):
      return 'phone number'
    elif re.search(r'Landroid/bluetooth/BluetoothAdapter;->getAddress\(\)Ljava/lang/String;', x):
      return 'L2 address (Bluetooth)'
    elif re.search(r'Landroid/net/wifi/WifiInfo;->getMacAddress\(\)Ljava/lang/String;|Ljava/net/NetworkInterface;->getHardwareAddress\(\)', x):
      return 'L2 address (Wi-Fi)'
    return None

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for op in q.invocations(InvocationPattern('invoke-', r'Landroid/provider/Settings\$Secure;->getString\(Landroid/content/ContentResolver;Ljava/lang/String;\)Ljava/lang/String;|Landroid/telephony/TelephonyManager;->getDeviceId\(\)Ljava/lang/String;|Landroid/telephony/TelephonyManager;->getSubscriberId\(\)Ljava/lang/String;|Landroid/telephony/TelephonyManager;->getLine1Number\(\)Ljava/lang/String;|Landroid/bluetooth/BluetoothAdapter;->getAddress\(\)Ljava/lang/String;|Landroid/net/wifi/WifiInfo;->getMacAddress\(\)Ljava/lang/String;|Ljava/net/NetworkInterface;->getHardwareAddress\(\)')):
      qn = q.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      val_type = self.analyzed(q, op)
      if val_type is not None:
        self._helper.raise_issue(self._helper.build_issue(
          sigid=self._id,
          cfd='certain',
          cvss=self._cvss,
          title=self._summary,
          info0=f'getting {val_type}',
          aff0=q.qualname_of(op.addr)
        ))

class PrivacySMSDetector(SignatureMixin):
  _id = 'privacy-sms'
  _cvss = 'CVSS:3.0/AV:N/AC:H/PR:L/UI:R/S:C/C:L/I:N/A:N/'
  _summary = 'privacy concerns'

  @staticmethod
  def create(helper: SignatureHelper) -> Signature:
    return PrivacySMSDetector(helper)

  def get_sigs(self) -> SignatureMap:
    return {self._id:dict(e=self.detect, d='Detects SMS-related behavior')}

  async def detect(self) -> None:
    context = self._get_context()
    store = context.store()
    q = store.query()
    for op in q.invocations(InvocationPattern('invoke-', r'Landroid/net/Uri;->parse\(Ljava/lang/String;\)Landroid/net/Uri;')):
      qn = q.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      try:
        if DataFlow(q).solved_constant_data_in_invocation(op, 0).startswith('content://sms/'):
          self._helper.raise_issue(self._helper.build_issue(
            sigid=self._id,
            cfd='certain',
            cvss=self._cvss,
            title=self._summary,
            info0='accessing SMS',
            aff0=q.qualname_of(op.addr)
          ))
      except DataFlow.NoSuchValueError:
        pass

    for op in q.invocations(InvocationPattern('invoke-', r'Landroid/telephony/SmsManager;->send')):
      qn = q.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cfd='certain',
        cvss=self._cvss,
        title=self._summary,
        info0='sending SMS',
        aff0=q.qualname_of(op.addr)
      ))

    for op in q.invocations(InvocationPattern('invoke-', r'Landroid/telephony/SmsMessage;->createFromPdu\(')):
      qn = q.qualname_of(op.addr)
      if context.is_qualname_excluded(qn):
        continue
      self._helper.raise_issue(self._helper.build_issue(
        sigid=self._id,
        cvss=self._cvss,
        title=self._summary,
        info0='intercepting incoming SMS',
        aff0=q.qualname_of(op.addr)
      ))



from __future__ import annotations
from typing import TYPE_CHECKING

import os.path
import re
from trueseeing.core.db import Query
from trueseeing.core.z import zd

if TYPE_CHECKING:
  from typing import Any, AnyStr, Final
  from sqlite3 import Connection

class Store:
  _fn: Final[str] = 'store.db'
  db: Connection

  def __init__(self, path: str) -> None:
    self._path = path
    self.db = self._open_db()
    self._check_schema()

  def invalidate(self) -> None:
    if self.db:
      self.db.close()
    self.db = self._open_db()
    self._check_schema()

  def _open_db(self) -> Connection:
    import sqlite3
    store_path = os.path.join(self._path, self._fn)
    is_creating = not os.path.exists(store_path)
    o = sqlite3.connect(store_path)
    o.create_function("REGEXP", 2, self._re_fn, deterministic=True)
    o.create_function("MZMATCHES", 3, self._mzmatches_fn, deterministic=True)
    self._prep_schema(o, is_creating)
    return o

  def _prep_schema(self, o: Connection, is_creating: bool) -> None:
    from trueseeing.core.db import StorePrep, FileTablePrep
    StorePrep(o).stage0()
    if is_creating:
      FileTablePrep(o).prepare()
      StorePrep(o).stage1()

  def _check_schema(self) -> None:
    from trueseeing.core.db import StorePrep
    StorePrep(self.db).require_valid_schema()

  @classmethod
  def require_valid_schema_on(cls, path: str) -> None:
    import os.path
    import sqlite3
    from trueseeing.core.db import StorePrep
    store_path = os.path.join(path, cls._fn)
    if not os.path.exists(store_path):
      from trueseeing.core.exc import InvalidSchemaError
      raise InvalidSchemaError()
    else:
      o = sqlite3.connect(store_path)
      StorePrep(o).require_valid_schema()

  @staticmethod
  def _re_fn(expr: AnyStr, item: Any) -> bool:
    if item is not None:
      return re.compile(expr).search(item) is not None
    else:
      return False

  @staticmethod
  def _mzmatches_fn(z: bool, expr: bytes, item: bytes) -> bool:
    if item is not None:
      return re.compile(expr).search(zd(item) if z else item) is not None
    else:
      return False

  def query(self) -> Query:
    return Query(store=self)



from __future__ import annotations
from typing import TYPE_CHECKING, NamedTuple

from trueseeing.core.cvss import CVSS3Scoring

if TYPE_CHECKING:
  from typing import Optional
  from typing_extensions import Literal
  IssueSeverity = Literal['critical', 'high', 'medium', 'low', 'info']
  IssueConfidence = Literal['certain', 'firm', 'tentative']

class Issue(NamedTuple):
  sigid: str
  cvss: str
  title: str
  cfd: IssueConfidence = 'firm'
  summary: Optional[str] = None
  desc: Optional[str] = None
  ref: Optional[str] = None
  sol: Optional[str] = None
  info0: Optional[str] = None
  info1: Optional[str] = None
  info2: Optional[str] = None
  aff0: Optional[str] = None
  aff1: Optional[str] = None
  aff2: Optional[str] = None

  @property
  def sev(self) -> IssueSeverity:
    return CVSS3Scoring.severity_of(self.score)

  @property
  def score(self) -> float:
    return CVSS3Scoring.score_of(self.cvss)

  def brief_desc(self) -> str:
    return ': '.join(filter(None, (self.title, self.brief_info())))

  def brief_info(self) -> str:
    return ': '.join(filter(None, (self.info0, self.info1, self.info2)))



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import List
  from trueseeing.api import CommandHelper, Command, CommandMap, ModifierMap

class ScanCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ScanCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '?s?':dict(e=self._help_signature, n='?s?', d='signature help'),
      'as':dict(e=self._scan, n='as[!]', d='run a scan (!: clear current issues)'),
      'as!':dict(e=self._scan),
    }

  def get_modifiers(self) -> ModifierMap:
    return {
      's':dict(n='@s:sig', d='include sig', e=None),
      'x':dict(n='@x:pa.ckage.name', d='exclude package', e=None),
    }

  async def _help_signature(self, args: deque[str]) -> None:
    from trueseeing.core.scan import Scanner
    ui.success('Signatures:')
    sigs = Scanner.get_all_signatures()
    width = 2 + max([len(k) for k in sigs.keys()])
    for k in sorted(sigs.keys()):
      ui.stderr(
        f'{{n:<{width}s}}{{d}}'.format(n=k, d=sigs[k]['d'])
      )

  async def _scan(self, args: deque[str]) -> None:
    import time
    from trueseeing.core.scan import Scanner
    from trueseeing.core.ui import ScanProgressReporter

    self._helper.require_target()

    cmd = args.popleft()

    context = await self._helper.get_context().analyze()
    limit = self._helper.get_graph_size_limit(self._helper.get_modifiers(args))
    sigsels = self._get_sigsels(self._helper.get_modifiers(args))
    try:
      scanner = Scanner(context, sigsels=sigsels, max_graph_size=limit)
    except ValueError as e:
      ui.fatal('invalid sigsels found: {}'.format(', '.join(sorted(e.args[0]))))

    if not scanner.get_active_signatures():
      ui.fatal('no signature selected')

    at = time.time()
    with context.store().query().scoped() as q:
      if cmd.endswith('!'):
        ui.info('clearing current issues')
        await scanner.clear(q)

      with ScanProgressReporter().scoped():
        nr = await scanner.scan(q)

      ui.success("done, found {nr} issues ({t:.02f} sec.)".format(nr=nr, t=(time.time() - at)))

  def _get_sigsels(self, mods: List[str]) -> List[str]:
    for m in mods:
      if m.startswith('@s:'):
        return m[3:].split(',')
    return []



from __future__ import annotations
from typing import TYPE_CHECKING

from contextlib import contextmanager
from trueseeing.core.model.issue import Issue
from trueseeing.core.tools import noneif
from trueseeing.core.z import ze, zd

if TYPE_CHECKING:
  from typing import Any, Iterable, Tuple, Optional, Iterator, List, TypedDict
  from typing_extensions import Self
  from sqlite3 import Connection
  from trueseeing.core.store import Store
  from trueseeing.core.model.issue import IssueConfidence

  class FileEntry(TypedDict):
    path: str
    blob: bytes
    z: bool

class StorePrep:
  def __init__(self, c: Connection) -> None:
    self.c = c

  def stage0(self) -> None:
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'store.s.sql').read_text())

  def stage1(self) -> None:
    from importlib.resources import files
    self.c.execute('pragma user_version={}'.format(self._get_cache_schema_id()))
    self.c.executescript((files('trueseeing')/'libs'/'store.0.sql').read_text())

  def require_valid_schema(self) -> None:
    v, = self.c.execute('pragma user_version').fetchone()
    if v != self._get_cache_schema_id():
      from trueseeing.core.exc import InvalidSchemaError
      raise InvalidSchemaError()

  def _get_cache_schema_id(self) -> int:
    from trueseeing.core.env import get_cache_schema_id
    return get_cache_schema_id()

class FileTablePrep:
  def __init__(self, c: Connection) -> None:
    self.c = c

  def prepare(self) -> None:
    from importlib.resources import files
    self.c.executescript((files('trueseeing')/'libs'/'files.0.sql').read_text())

class Query:
  def __init__(self, store: Store) -> None:
    self.db = store.db

  @contextmanager
  def scoped(self) -> Iterator[Self]:
    with self.db:
      yield self

  def file_find(self, pat: str, regex: bool = False) -> Iterable[str]:
    for f, in self.db.execute('select path from files where path {op} :pat'.format(op=('like' if not regex else 'regexp')), dict(pat=pat)):
      yield f

  def file_get(self, path: str, default: Optional[bytes] = None, patched: bool = False) -> Optional[bytes]:
    b: bytes
    stmt0 = 'select z, blob from files where path=:path'
    stmt1 = 'select A.z as z,coalesce(B.blob, A.blob) as blob from files as A full outer join patches as B using (path) where path=:path'
    for z, b in self.db.execute(stmt1 if patched else stmt0, dict(path=path)):
      return zd(b) if z else b
    else:
      return default

  def file_get_xml(self, path: str, default: Any = None, patched: bool = False) -> Any:
    import lxml.etree as ET
    r = self.file_get(path, patched=patched)
    if r is not None:
      return ET.fromstring(r, parser=ET.XMLParser(recover=True))
    else:
      return default

  def file_enum(self, pat: Optional[str], patched: bool = False, regex: bool = False, neg: bool = False) -> Iterable[Tuple[str, bytes]]:
    if pat is not None:
      stmt0 = 'select path, z, blob from files where {neg} path {op} :pat'.format(neg='not' if neg else '', op=('like' if not regex else 'regexp'))
      stmt1 = 'select path, coalesce(B.z, A.z) as z, coalesce(B.blob, A.blob) as blob from files as A full outer join patches as B using (path) where {neg} path {op} :pat'.format(neg='not' if neg else '', op=('like' if not regex else 'regexp'))
      for n, z, o in self.db.execute(stmt1 if patched else stmt0, dict(pat=pat)):
        yield n, zd(o) if z else o
    else:
      stmt2 = 'select path, z, blob from files'
      stmt3 = 'select path, coalesce(B.z, A.z) as z, coalesce(B.blob, A.blob) from files as A full outer join patches as B using (path)'
      for n, z, o in self.db.execute(stmt3 if patched else stmt2):
        yield n, zd(o) if z else o

  def file_count(self, pat: Optional[str], patched: bool = False, regex: bool = False, neg: bool = False) -> int:
    if pat is not None:
      stmt0 = 'select count(1) from files where {neg} path {op} :pat'.format(neg='not' if neg else '', op=('like' if not regex else 'regexp'))
      stmt1 = 'select conut(1) from files as A full outer join patches as B using (path) where {neg} path {op} :pat'.format(neg='not' if neg else '', op=('like' if not regex else 'regexp'))
      for nr, in self.db.execute(stmt1 if patched else stmt0, dict(pat=pat)):
        return nr # type:ignore[no-any-return]
    else:
      stmt2 = 'select count(1) from files'
      stmt3 = 'select count(1) from files as A full outer join patches as B using (path)'
      for nr, in self.db.execute(stmt3 if patched else stmt2):
        return nr # type:ignore[no-any-return]
    return 0

  def file_put_batch(self, gen: Iterable[FileEntry]) -> None:
    self.db.executemany(
      'insert into files (path, blob, z) values (:path,:blob,:z)',
      (dict(path=e['path'], blob=ze(e['blob']) if e['z'] else e['blob'], z=e['z']) for e in gen)
    )

  def patch_enum(self, pat: Optional[str]) -> Iterable[Tuple[str, bytes]]:
    if pat is not None:
      stmt0 = 'select path, z, blob from patches where path like :pat'
      for n, z, o in self.db.execute(stmt0, dict(pat=pat)):
        yield n, zd(o) if z else o
    else:
      stmt1 = 'select path, z, blob from patches'
      for n, z, o in self.db.execute(stmt1):
        yield n, zd(o) if z else o

  def patch_put(self, path: str, blob: bytes, z: bool) -> None:
    self.db.execute(
      'replace into patches (path, blob, z) values (:path,:blob,:z)',
      dict(path=path, blob=ze(blob) if z else blob, z=z)
    )

  def patch_exists(self, path: Optional[str]) -> bool:
    stmt0 = 'select 1 from patches where path=:path'
    stmt1 = 'select 1 from patches'
    for r, in self.db.execute(stmt0 if path is not None else stmt1, dict(path=path)):
      return True
    else:
      return False

  def patch_clear(self) -> None:
    self.db.execute('delete from patches')

  def issue_count(self) -> int:
    for nr, in self.db.execute('select count(1) from analysis_issues'):
      return int(nr)
    else:
      return 0

  def issue_raise(self, i: Issue) -> None:
    assert i.score is not None
    self.db.execute(
      'insert or ignore into analysis_issues (sig, title, summary, descr, ref, sol, info0, info1, info2, cfd, score, cvss, aff0, aff1, aff2) values (:sigid, :title, :summary, :desc, :ref, :sol, :info0, :info1, :info2, :cfd, :score, :cvss, :aff0, :aff1, :aff2)',
      dict(
        sigid=i.sigid,
        title=i.title,
        cfd=self._issue_confidence_to_int(i.cfd),
        cvss=i.cvss,
        score=i.score,
        summary=noneif(i.summary, ''),
        desc=noneif(i.desc, ''),
        ref=noneif(i.ref, ''),
        sol=noneif(i.sol, ''),
        info0=noneif(i.info0, ''),
        info1=noneif(i.info1, ''),
        info2=noneif(i.info2, ''),
        aff0=noneif(i.aff0, ''),
        aff1=noneif(i.aff1, ''),
        aff2=noneif(i.aff2, ''),
      ))

  def issue_clear(self) -> None:
    self.db.execute('delete from analysis_issues')

  def issues(self) -> Iterable[Issue]:
    for m in self.db.execute('select sig, title, summary, descr, ref, sol, info0, info1, info2, cfd, score, cvss, aff0, aff1, aff2 from analysis_issues'):
      yield self._issue_from_row(m)

  def findings_list(self) -> Iterable[Tuple[int, Tuple[str, str, Optional[str], Optional[str], Optional[str], Optional[str], float, str]]]:
    for no, r in enumerate(self.db.execute('select distinct sig, title, summary, descr, ref, sol, score, cvss from analysis_issues order by score desc')):
      yield no, (
        r[0],
        r[1],
        r[2] if r[2] else None,
        r[3] if r[3] else None,
        r[4] if r[4] else None,
        r[5] if r[5] else None,
        r[6],
        r[7],
      )

  def issues_by_group(self, *, sig: str, title: str) -> Iterable[Issue]:
    for m in self.db.execute('select sig, title, summary, descr, ref, sol, info0, info1, info2, cfd, score, cvss, aff0, aff1, aff2 from analysis_issues where sig=:sig and title=:title order by score desc, cfd desc', dict(sig=sig, title=title)):
      yield self._issue_from_row(m)

  @classmethod
  def _issue_confidence_to_int(cls, c: IssueConfidence) -> int:
    return dict(certain=2, firm=1, tentative=0)[c]

  @classmethod
  def _issue_confidence_from_int(cls, c: int) -> IssueConfidence:
    m: List[IssueConfidence] = ['tentative', 'firm', 'certain']
    return m[c]

  @classmethod
  def _issue_from_row(cls, r: Tuple[Any, ...]) -> Issue:
    return Issue(
      sigid=r[0],
      title=r[1],
      cfd=cls._issue_confidence_from_int(r[9]),
      cvss=r[11],
      summary=r[2] if r[2] else None,
      desc=r[3] if r[3] else None,
      ref=r[4] if r[4] else None,
      sol=r[5] if r[5] else None,
      info0=r[6] if r[6] else None,
      info1=r[7] if r[7] else None,
      info2=r[8] if r[8] else None,
      aff0=r[12] if r[12] else None,
      aff1=r[13] if r[13] else None,
      aff2=r[14] if r[14] else None,
    )



from __future__ import annotations
from typing import TYPE_CHECKING

from trueseeing.core.store import Store
from trueseeing.core.android.db import APKQuery

if TYPE_CHECKING:
  from sqlite3 import Connection

class APKStore(Store):
  def _prep_schema(self, o: Connection, is_creating: bool) -> None:
    from trueseeing.core.db import FileTablePrep
    from trueseeing.core.android.db import APKStorePrep
    APKStorePrep(o).stage0()
    if is_creating:
      FileTablePrep(o).prepare()
      APKStorePrep(o).stage1()

  def _check_schema(self) -> None:
    from trueseeing.core.android.db import APKStorePrep
    APKStorePrep(self.db).require_valid_schema()

  @classmethod
  def require_valid_schema_on(cls, path: str) -> None:
    import os.path
    store_path = os.path.join(path, cls._fn)
    if not os.path.exists(store_path):
      from trueseeing.core.exc import InvalidSchemaError
      raise InvalidSchemaError()
    else:
      import sqlite3
      from trueseeing.core.android.db import APKStorePrep
      o = sqlite3.connect(store_path)
      APKStorePrep(o).require_valid_schema()

  def query(self) -> APKQuery:
    return APKQuery(store=self)



from __future__ import annotations
from typing import TYPE_CHECKING
from functools import cache
import os
import re

if TYPE_CHECKING:
    from typing import Iterable, Dict, Any, Iterator, Mapping, Tuple

def _analyzed(x: str, tlds: re.Pattern[str]) -> Iterable[Dict[str, Any]]:
    if '://' in x:
        yield dict(type_='URL', value=re.findall(r'\S+://\S+', x))
    elif re.search(r'^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+', x):
        yield dict(type_='path component', value=re.findall(r'^/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+', x))
    elif re.search(r'^[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)+(:[0-9]+)?$', x):
        m = re.search(r'^([^:/]+)', x)
        if m:
            hostlike = m.group(1)
            components = hostlike.split('.')
            if len(components) == 4 and all(re.match(r'^\d+$', c) for c in components) and all(int(c) < 256 for c in components):
                yield dict(type_='possible IPv4 address', value=[hostlike])
            elif tlds.search(components[-1]):
                if not re.search(r'^android\.(intent|media)\.|^os\.name$|^java\.vm\.name|^[A-Z]+.*\.(java|EC|name|secure)$', hostlike):
                    yield dict(type_='possible FQDN', value=[hostlike])

@cache
def _pat(c: str) -> re.Pattern[str]:
    from io import StringIO
    f = StringIO(c)
    return re.compile('^(?:{})$'.format('|'.join(re.escape(l.strip()) for l in f if l and not l.startswith('#'))), flags=re.IGNORECASE)

@cache
def _tlds() -> str:
    from importlib.resources import files
    with (files('trueseeing')/'libs'/'public_suffix_list.dat').open('r', encoding='utf-8') as f:
      return f.read()

def analyze_url(path: str) -> Iterator[Mapping[str, Any]]:
    def walker(path: str) -> Iterator[Tuple[str, bytes]]:
        for dirpath, _, filenames in os.walk(path):
            for fn in filenames:
                target = os.path.join(dirpath, fn)
                with open(target, 'rb') as f:
                    yield target, f.read()

    return analyze_url_in(walker(path))

def analyze_url_in(gen: Iterable[Tuple[str, bytes]]) -> Iterator[Mapping[str, Any]]:
    tlds = _pat(_tlds())

    pats = rb'|'.join([
        b"([a-z0-9A-Z]+://[^\\\"<>()' \\t\\n\\v\\r\\x00-\\x1f\\x80-\\xff]+)",
        rb'"/[{}$%a-zA-Z0-9_-]+(/[{}$%a-zA-Z0-9_-]+)+',
        rb'"[a-zA-Z0-9-]+(\.[a-zA-Z0-9-]+)+(:[0-9]+)?"',
    ])

    seen = set()
    for n, s in gen:
      for m in re.finditer(pats, s):
          urllike = m.group(0).decode('latin1').strip('"')
          if urllike not in seen:
              seen.add(urllike)
              for d in _analyzed(urllike, tlds):
                  for v in d['value']:
                      yield dict(fn=n, typ=d['type_'], v=v)



from __future__ import annotations
from typing import TYPE_CHECKING

from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Dict, Any
  from trueseeing.api import CommandHelper, Command, CommandMap, ConfigEntry, ModifierMap, ModifierEvent

class ConfigCommand(CommandMixin):
  _confbag: Dict[str, ConfigEntry]
  _stash: Dict[str, Any]

  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper
    self._confbag = self._helper._confbag  # type:ignore[attr-defined]
    self._stash = dict()

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ConfigCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      '?e?':dict(e=self._help, n='?e?', d='config help'),
      'e':dict(e=self._manip, n='e key[=value]', d='get/set config'),
    }

  def get_modifiers(self) -> ModifierMap:
    return {
      'e':dict(e=self._mod, n='@e:key=value', d='bind the config for the command'),
    }

  async def _help(self, args: deque[str]) -> None:
    ui.success('Configs:')
    if self._confbag:
      width = (2 + max([len(e.get('d', '')) for e in self._confbag.values()]) // 4) * 4
      for k in sorted(self._confbag):
        e = self._confbag[k]
        if 'n' in e:
          ui.stderr(
            f'{{n:<{width}s}}{{d}}'.format(n=e['n'], d=e['d'])
          )

  async def _manip(self, args: deque[str]) -> None:
    from trueseeing.core.exc import InvalidConfigKeyError
    _ = args.popleft()
    if not args:
      ui.fatal('need a config key')
    kv = args.popleft()
    if args:
      ui.fatal('got an unexpected token (try e key=value to set)')
    try:
      if '=' not in kv:
        key = kv
        ui.info('{}: {}'.format(key, self._helper.get_config(key)))
      else:
        key, value = kv.split('=')
        if not value:
          ui.fatal('need a value')
        self._helper.set_config(key, value)
    except InvalidConfigKeyError:
      ui.fatal(f'unknown key: {key}')

  async def _mod(self, ev: ModifierEvent, val: str) -> None:
    from trueseeing.core.exc import InvalidConfigKeyError
    if '=' not in val:
      ui.fatal('need a value')
    try:
      k, v = val.split('=')
      if ev == 'begin':
        if k in self._stash:
          ui.fatal(f'config key is already bound: {k}')
        ov = self._helper.get_config(k)
        self._helper.set_config(k, v)
        self._stash[k] = ov
      elif ev == 'end':
        assert k in self._stash
        try:
          self._helper.set_config(k, self._stash[k])
        finally:
          del self._stash[k]
    except InvalidConfigKeyError:
      ui.fatal(f'unknown key: {k}')



from __future__ import annotations
from typing import TYPE_CHECKING

from contextlib import contextmanager

if TYPE_CHECKING:
  from typing import List, Optional, Iterator, Dict, Any, Set
  from trueseeing.api import SignatureEntry, SignatureHelper, SignatureMap
  from trueseeing.core.context import Context, ContextType
  from trueseeing.core.db import Query
  from trueseeing.core.model.issue import Issue, IssueConfidence

class Scanner:
  _helper: SignatureHelper
  _sigs: Dict[str, SignatureEntry]

  def __init__(self, context: Context, *, sigsels: List[str] = [], excludes: List[str] = [], max_graph_size: Optional[int] = None) -> None:
    from trueseeing.core.config import Configs
    self._context = context
    self._sigs = dict()
    self._excludes = excludes
    self._max_graph_size = max_graph_size
    self._confbag = Configs.get().bag
    self._helper = SignatureHelperImpl(self)

    self._init_sigs(['all'] + sigsels)

  def get_active_signatures(self) -> SignatureMap:
    return self._sigs

  @classmethod
  def get_all_signatures(cls) -> SignatureMap:
    return Scanner(context=None).get_active_signatures()  # type: ignore[arg-type]

  async def scan(self, q: Query) -> int:
    import asyncio
    from pubsub import pub
    from trueseeing.core.exc import InvalidContextError
    from trueseeing.core.android.analyze.flow import DataFlow
    from trueseeing.core.ui import ui
    with DataFlow.apply_max_graph_size(self._max_graph_size):
      with self._apply_excludes_on_context():
        def _detected(issue: Issue) -> None:
          q.issue_raise(issue)

        async def _call(id_: str, ent: SignatureEntry) -> None:
          try:
            await ent['e']()
          except InvalidContextError:
            ui.warn(f'scan: {id_}: context invalid, signature ignored')

        pub.subscribe(_detected, 'issue')
        await asyncio.gather(*[_call(k, v) for k,v in self._sigs.items()])
        pub.unsubscribe(_detected, 'issue')

        return q.issue_count()

  async def clear(self, q: Query) -> None:
    q.issue_clear()

  @contextmanager
  def _apply_excludes_on_context(self) -> Iterator[None]:
    o = self._context.excludes
    self._context.excludes = self._excludes
    yield None
    self._context.excludes = o

  @classmethod
  def _sigsel_matches(cls, sigid: str, sels: List[str]) -> bool:
    def _match(sigid: str, sel: str) -> Optional[bool]:
      neg = False
      if sel.startswith('no-'):
        sel = sel[3:]
        neg = True
      if sel == 'all':
        return not neg
      elif sel.endswith('-all'):
        psel = sel[:-4]
        if sigid.startswith(psel):
          return not neg
        else:
          return None
      else:
        if sigid == sel:
          return not neg
        else:
          return None
    for o in (_match(sigid, x) for x in reversed(sels)):
      if o is None:
        continue
      return o
    else:
      return False

  def _require_sigsels_fully_applied(self, sigsels: List[str], known: Set[str]) -> None:
    def _regen(key: str, neg: bool) -> str:
      return ('no-' if neg else '') + key
    unknown = set()
    for sel in set(sigsels):
      neg = False
      if sel.startswith('no-'):
        sel = sel[3:]
        neg = True
      if sel != 'all':
        if sel.endswith('-all'):
          psel = sel[:-4]
          if psel and not any((f'{psel}-' in k) for k in known):
            unknown.add(_regen(sel, neg))
        elif sel not in known:
          unknown.add(_regen(sel, neg))
    if unknown:
      raise ValueError(unknown)

  def _init_sigs(self, sigsels: List[str]) -> None:
    from itertools import chain
    from trueseeing.sig import discover
    from trueseeing.core.ext import Extension
    known = set()
    for clazz in chain(discover(), Extension.get().get_signatures()):
      matched = False
      t = clazz.create(self._helper)
      for k,v in t.get_sigs().items():
        known.add(k)
        if self._sigsel_matches(k, sigsels):
          self._sigs[k] = v
          matched = True
      if matched:
        self._confbag.update(t.get_configs())
    self._require_sigsels_fully_applied(sigsels, known)

class SignatureHelperImpl:
  def __init__(self, scanner: Scanner) -> None:
    self._s = scanner
    self._confbag = self._s._confbag
  def get_context(self, typ: Optional[ContextType] = None) -> Any:
    if typ:
      from trueseeing.core.ui import ui
      ui.warn('get_context(typ): deprecated, use get_context().request_type(...)', onetime=True)
      self._s._context.require_type(typ)
    return self._s._context
  def raise_issue(self, issue: Issue) -> None:
    from pubsub import pub
    pub.sendMessage('issue', issue=issue)
  def build_issue(
      self,
      sigid: str,
      cvss: str,
      title: str,
      cfd: IssueConfidence = 'firm',
      summary: Optional[str] = None,
      desc: Optional[str] = None,
      ref: Optional[str] = None,
      sol: Optional[str] = None,
      info0: Optional[str] = None,
      info1: Optional[str] = None,
      info2: Optional[str] = None,
      aff0: Optional[str] = None,
      aff1: Optional[str] = None,
      aff2: Optional[str] = None,
  ) -> Issue:
    from trueseeing.core.model.issue import Issue
    return Issue(
      sigid=sigid,
      cvss=cvss,
      cfd=cfd,
      title=title,
      desc=desc,
      ref=ref,
      sol=sol,
      summary=summary,
      info0=info0,
      info1=info1,
      info2=info2,
      aff0=aff0,
      aff1=aff1,
      aff2=aff2,
    )
  def get_config(self, k: str) -> Any:
    from trueseeing.core.exc import InvalidConfigKeyError
    try:
      e = self._confbag[k]
    except KeyError:
      raise InvalidConfigKeyError()
    else:
      return e['g']()

  def set_config(self, k: str, v: Any) -> None:
    from trueseeing.core.exc import InvalidConfigKeyError
    try:
      e = self._confbag[k]
    except KeyError:
      raise InvalidConfigKeyError()
    else:
      e['s'](v)



from __future__ import annotations
from typing import TYPE_CHECKING

import sys
from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Optional
  from trueseeing.api import CommandHelper, Command, CommandMap

class ShowCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ShowCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'pf':dict(e=self._show_file, n='pf[x][!] path [output.bin]', d='show file (x: hex)'),
      'pf!':dict(e=self._show_file),
      'pfx':dict(e=self._show_file),
      'pfx!':dict(e=self._show_file),
    }

  async def _show_file(self, args: deque[str]) -> None:
    outfn: Optional[str] = None

    self._helper.require_target()

    cmd = args.popleft()

    if not args:
      ui.fatal('need path')

    path = args.popleft()

    if args:
      import os
      outfn = args.popleft()
      if os.path.exists(outfn) and not cmd.endswith('!'):
        ui.fatal('outfile exists; force (!) to overwrite')

    from binascii import hexlify

    context = await self._helper.get_context().analyze(level=1)
    level = context.get_analysis_level()
    if level < 3:
      ui.warn('detected analysis level: {} ({}) -- try analyzing fully (\'aa\') to maximize coverage'.format(level, self._helper.decode_analysis_level(level)))
    d = context.store().query().file_get(path)
    if d is None:
      ui.fatal('file not found')
    if outfn is None:
      sys.stdout.buffer.write(d if 'x' not in cmd else hexlify(d))
    else:
      with open(outfn, 'wb') as f:
        f.write(d if 'x' not in cmd else hexlify(d))



from __future__ import annotations
from typing import TYPE_CHECKING

import os

from trueseeing.core.cvss import CVSS3Scoring
from trueseeing.core.tools import noneif
from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import List, Protocol, Any, Dict, TextIO, Set
  from typing_extensions import Literal
  from trueseeing.core.context import Context
  from trueseeing.core.model.issue import Issue

  ReportFormat = Literal['html', 'json']

  class ReportGenerator(Protocol):
    def __init__(self, context: Context) -> None: ...
    def generate(self, f: TextIO) -> None: ...

class ConsoleNoter:
  _seen: Set[Issue]

  def __init__(self) -> None:
    self._seen = set()

  def note(self, issue: Issue) -> None:
    if issue not in self._seen:
      ui.info(self._formatted(issue))
      self._seen.add(issue)

  @classmethod
  def _formatted(cls, issue: Issue) -> str:
    return '{aff0}:{aff1}:{aff2}:{sev}{{{cfd}}}:{desc} [{sigid}]'.format(
      aff0=noneif(issue.aff0, '(global)'),
      aff1=noneif(issue.aff1, 0),
      aff2=noneif(issue.aff2, 0),
      sev=issue.sev,
      cfd=issue.cfd,
      desc=issue.brief_desc(),
      sigid=issue.sigid
    )

class CIReportGenerator:
  def __init__(self, context: Context) -> None:
    self._context = context

  def generate(self, f: TextIO) -> None:
    for issue in self._context.store().query().issues():
      f.write(ConsoleNoter._formatted(issue) + '\n')

class HTMLReportGenerator:
  def __init__(self, context: Context) -> None:
    from trueseeing import __version__
    self._context = context
    self._toolchain = dict(version=__version__)

  def generate(self, f: TextIO) -> None:
    with self._context.store().query().scoped() as query:
      from datetime import datetime
      from zoneinfo import ZoneInfo

      app: Dict[str, Any] = dict(
        path=self._context.target,
      )
      issues = []

      ts = datetime.now(tz=ZoneInfo('UTC')).isoformat(timespec='seconds')

      if 'apk' in self._context.type:
        ctx = self._context.require_type('apk')
        apk = ctx.target
        manif = ctx.parsed_manifest()
        ns = dict(android='http://schemas.android.com/apk/res/android')

        app.update(dict(
          package=manif.xpath('/manifest/@package', namespaces=ns)[0],
          version_name=manif.xpath('/manifest/@android:versionName', namespaces=ns)[0],
          version_code=manif.xpath('/manifest/@android:versionCode', namespaces=ns)[0],
          size=os.stat(apk).st_size,
        ))

      for no, row in query.findings_list():
        instances: List[Dict[str, Any]] = []
        issues.append(dict(no=no, sig=row[0], title=row[1].title(), summary=row[2], desc=row[3], ref=row[4], sol=row[5], score=row[6], cvss=row[7], sev=CVSS3Scoring.severity_of(row[6]).title(), insts=instances, severity_panel_style={'critical':'panel-danger', 'high':'panel-warning', 'medium':'panel-warning', 'low':'panel-success', 'info':'panel-info'}[CVSS3Scoring.severity_of(row[6])]))
        for issue in query.issues_by_group(sig=row[0], title=row[1]):
          instances.append(dict(info=issue.brief_info(), aff0=issue.aff0, aff1=issue.aff1, aff2=issue.aff2))

      app.update(dict(
        fp=self._context.fingerprint_of(),
        issues=len(issues),
        issues_critical=len([_ for _ in issues if _['sev'] == 'Critical']),
        issues_high=len([_ for _ in issues if _['sev'] == 'High']),
        issues_medium=len([_ for _ in issues if _['sev'] == 'Medium']),
        issues_low=len([_ for _ in issues if _['sev'] == 'Low']),
        issues_info=len([_ for _ in issues if _['sev'] == 'Info']),
      ))

      from importlib.resources import as_file, files
      from jinja2 import Environment, FileSystemLoader
      with as_file(files('trueseeing')/'libs'/'template') as path:
        env = Environment(loader=FileSystemLoader(path), autoescape=True, trim_blocks=True, lstrip_blocks=True)
        env.filters['excerpt'] = self._excerpt
        f.write(env.get_template('report.html').render(app=app, issues=issues, toolchain=self._toolchain, ts=ts))

  @staticmethod
  def _excerpt(x: str, w: int = 256, p: float = 0.214, ellipsis: str = ' ... ') -> str:
    l = len(x)
    if l < w:
      return x
    else:
      pl1 = int(w * p)
      pl0 = w - pl1 - len(ellipsis)
      return x[:pl0] + ellipsis + x[-pl1:]

class JSONReportGenerator:
  def __init__(self, context: Context) -> None:
    self._context = context

  def generate(self, f: TextIO) -> None:
    from json import dumps
    with self._context.store().query().scoped() as query:
      issues = []
      app: Dict[str, Any] = dict(
        path=self._context.target,
      )

      if 'apk' in self._context.type:
        ctx = self._context.require_type('apk')
        apk = ctx.target
        manif = ctx.parsed_manifest()
        ns = dict(android='http://schemas.android.com/apk/res/android')

        app.update(dict(
          package=manif.xpath('/manifest/@package', namespaces=ns)[0],
          version_name=manif.xpath('/manifest/@android:versionName', namespaces=ns)[0],
          version_code=manif.xpath('/manifest/@android:versionCode', namespaces=ns)[0],
          size=os.stat(apk).st_size,
        ))

      for no, row in query.findings_list():
        instances: List[Dict[str, Any]] = []
        issues.append(dict(
          no=no,
          sig=row[0],
          summary=row[1].title(),
          synopsis=row[2],
          description=row[3],
          seealso=row[4],
          solution=row[5],
          cvss3_score=row[6],
          cvss3_vector=row[7],
          severity=CVSS3Scoring.severity_of(row[6]).title(),
          instances=instances
          ))
        for issue in query.issues_by_group(sig=row[0], title=row[1]):
          instances.append(dict(
            info=issue.brief_info(),
            source=issue.aff0,
            row=issue.aff1,
            col=issue.aff2))

      app.update(dict(
        fp=self._context.fingerprint_of(),
        issues=len(issues)
      ))
      f.write(dumps({"app": app, "issues": issues}, indent=2))



from __future__ import annotations
from typing import TYPE_CHECKING

import os
import os.path

from pubsub import pub

from trueseeing.core.context import Context, Fingerprint
from trueseeing.core.env import get_cache_dir
from trueseeing.core.ui import ui
from trueseeing.core.ios.store import IPAStore

if TYPE_CHECKING:
  from typing import List, Optional, Final, Set, TypedDict, Iterator, Any, Mapping
  from trueseeing.core.context import ContextType
  from trueseeing.core.db import FileEntry
  from trueseeing.core.ios.db import IPAQuery

  class Call(TypedDict):
    path: str
    sect: str
    offs: int
    priv: bool
    swift: bool
    objc: bool
    cpp: bool
    target: str

class IPAContext(Context):
  wd: str
  excludes: List[str]
  _path: str
  _store: Optional['IPAStore'] = None
  _type: Final[Set[ContextType]] = {'ipa', 'file'}
  _fp = Fingerprint()

  def invalidate(self) -> None:
    super().invalidate()
    self._fp.get.cache_clear()

  def _get_type(self) -> Set[ContextType]:
    return self._type

  def _get_workdir(self) -> str:
    return os.path.join(get_cache_dir(), 'ts2-ios-{}'.format(self._get_fingerprint()))

  def _get_size(self) -> Optional[int]:
    return os.stat(self._path).st_size

  def _get_fingerprint(self) -> str:
    return self._fp.get(self._path)

  async def _recheck_schema(self) -> None:
    pass

  def store(self) -> 'IPAStore':
    if self._store is None:
      self._store = IPAStore(self.wd)
    return self._store

  async def _analyze(self, level: int) -> None:
    q: IPAQuery
    if level > 0:
      import plistlib
      with self.store().query().scoped() as q:
        from zipfile import ZipFile
        with ZipFile(self._path, 'r') as zf:
          def _decode(n: str, b: bytes) -> FileEntry:
            if not n.endswith('Info,plist'):
              return dict(path=n, blob=b, z=True)
            else:
              return dict(path=n, blob=plistlib.dumps(plistlib.loads(b)), z=True)
          q.file_put_batch(_decode(i.filename, zf.read(i)) for i in zf.infolist() if not i.is_dir())

    if level > 2:
      tarpath = os.path.join(os.path.dirname(self._path), 'disasm.tar.gz')
      if not os.path.exists(tarpath):
        ui.fatal(f'prepare {tarpath}')
      with self.store().query().scoped() as q:
        pub.sendMessage('progress.core.analysis.nat.begin')
        import tarfile
        with tarfile.open(tarpath) as tf:
          q.file_put_batch(dict(path=f'disasm/{i.name}', blob=tf.extractfile(i).read(), z=True) for i in tf.getmembers() if (i.isreg() or i.islnk())) # type:ignore[union-attr]

        if level > 3:
          pub.sendMessage('progress.core.analysis.nat.analyzing')
          from trueseeing.core.ios.analyze import analyze_api_in

          def _as_call(g: Iterator[Mapping[str, Any]]) -> Iterator[Call]:
            for e in g:
              typ = e['typ']
              lang = e['lang']
              sect, offs = e['origin'].split('+')
              yield dict(
                path=e['fn'],
                sect=sect,
                offs=int(offs.strip(), 16),
                priv=(typ == 'private'),
                swift=(lang == 'swift'),
                objc=(lang == 'objc'),
                cpp=(lang == 'cpp'),
                target=e['call']
              )

          q.call_add_batch(_as_call(analyze_api_in(q.file_enum('disasm/%'))))
          pub.sendMessage('progress.core.analysis.nat.summary', calls=q.call_count())



from __future__ import annotations
from typing import TYPE_CHECKING
import asyncio
import sys

from trueseeing.core.ui import ui

if TYPE_CHECKING:
  from typing import Optional, Coroutine, Any, Literal

  OpMode = Optional[Literal['scan', 'inspect', 'batch']]

class Shell:
  @classmethod
  def _version(cls) -> str:
    from trueseeing import __version__
    #   ..............................................................................80
    return (
      f'Trueseeing {__version__}, an non-decompiling Android app vulnerability scanner\n'
       'Copyright (C) Takahiro Yoshimura <alterakey@protonmail.com> et al.\n' # noqa: E131
       'All rights reserved.  Licensed under the terms of GNU General Public License Version 3 or later.\n'
    )

  def _launch(self, coro: Coroutine[Any, Any, int]) -> int:
    return asyncio.run(coro)

  def _deprecated(self, msg: str) -> None:
    ui.warn(f'warning: {msg}', onetime=True)

  def invoke(self) -> int:
    from argparse import ArgumentParser
    from trueseeing.core.ext import Extension

    log_level = ui.INFO
    cmdlines = []

    parser = ArgumentParser(description='Non-decompiling Android app vulnerability scanner')
    args_mut0 = parser.add_mutually_exclusive_group()
    args_mut1 = parser.add_mutually_exclusive_group()
    parser.add_argument('fn', nargs='?', metavar='FILE', help='Target APK file')
    parser.add_argument('--help-signatures', action='store_true', help='Show signatures')
    parser.add_argument('--help-formats', action='store_true', help='Show supported file formats')
    parser.add_argument('--version', action='store_true', help='Version information')
    parser.add_argument('--norc', action='store_true', help='Ignore startup file')
    parser.add_argument('--noext', action='store_true', help='Ignore extensions')
    parser.add_argument('--quiet', action='store_true', help='Be less verbose')
    parser.add_argument('-d', '--debug', action='store_true', help='Debug mode')
    parser.add_argument('-e', '--abort-on-errors', action='store_true', help='Abort on errors')
    parser.add_argument('-F', dest='force_opener', help='Open target as specified format')
    parser.add_argument('-n', dest='no_target', action='store_true', help='Run without target')
    args_mut0.add_argument('-i', dest='scriptfn', metavar='FILE', help='Run script file before prompt')
    args_mut0.add_argument('-c', dest='inline_cmd', metavar='COMMAND', help='Run commands before prompt')
    args_mut1.add_argument('-q', dest='mode', action='store_const', const='batch', help='Batch mode; quit instead of giving prompt')
    args_mut1.add_argument('-Q', dest='mode', action='store_const', const='batch_quiet', help='same as --quiet -q')
    args_mut1.add_argument('--inspect', dest='mode', action='store_const', const='inspect', help='Inspect mode (deprecated; now default)')
    args_mut1.add_argument('--scan', dest='mode', action='store_const', const='scan', help='Scan mode (deprecated; use -eqc "as;g*"; e.g. gh for HTML)')

    scan_args = parser.add_argument_group('Scan mode (DEPRECATED)')
    scan_args.add_argument('--scan-sigs', metavar='SIG,...', help='Select signatures (use --help-signatures to list signatures)')
    scan_args.add_argument('--scan-exclude', dest='scan_exclude_packages', action='append', metavar='PATTERN', help='Excluding packages matching pattern')
    scan_args.add_argument('--scan-output', dest='scan_output_filename', metavar='FILE', help='Report filename ("-" for stdout)')
    scan_args.add_argument('--scan-report', dest='scan_report_format', choices=['html', 'json'], help='Report format (html: HTML (default), json: JSON)')
    scan_args.add_argument('--scan-max-graph-size', type=int, metavar='N', help='Limit graph size')
    scan_args_mut = scan_args.add_mutually_exclusive_group()
    scan_args_mut.add_argument('--scan-update-cache', action='store_true', help='Analyze and rebuild codebase cache')
    scan_args_mut.add_argument('--scan-no-cache', action='store_true', help='Do not keep codebase cache')
    args = parser.parse_args()

    if args.mode == 'batch_quiet':
      args.mode = 'batch'
      args.quiet = True

    if args.debug:
      log_level = ui.DEBUG
    if args.quiet:
      log_level = ui.WARN
    if args.noext:
      ui.warn('disabling extensions')
      Extension.disabled = True

    if not args.mode:
      args.mode = 'inspect'
    elif args.mode == 'inspect':
      self._deprecated('--inspect is deprecated; ignored as default')
    elif args.mode == 'scan':
      self._deprecated('--scan is deprecated; use -eqc "as;g*"; e.g. gh for HTML')
    if args.inline_cmd:
      cmdlines = [args.inline_cmd]
    if args.scriptfn:
      try:
        with open(args.scriptfn, 'r') as f:
          cmdlines = [l for l in f]
      except OSError as e:
        ui.fatal(f'cannot open script file: {e}')
    if args.scan_max_graph_size:
      from trueseeing.core.android.analyze.flow import DataFlow
      DataFlow.set_max_graph_size(args.scan_max_graph_size)
    if args.version:
      ui.stderr(self._version())
      return 0
    if args.help_signatures:
      args.no_target = True
      args.mode = 'batch'
      cmdlines = ['?s?']
    if args.help_formats:
      args.no_target = True
      args.mode = 'batch'
      cmdlines = ['?f?']

    ui.set_level(log_level)

    if args.no_target:
      args.fn = None
    elif not args.fn:
      parser.print_help()
      return 2

    if args.mode in ['inspect', 'batch']:
      from trueseeing.app.inspect import InspectMode

      if not args.norc:
        from trueseeing.core.env import get_rc_path
        try:
          with open(get_rc_path(), 'r') as f:
            rc = [l for l in f]
          cmdlines = rc + cmdlines
        except FileNotFoundError:
          pass
        except OSError as e:
          ui.warn(f'cannot open rc file, ignoring: {e}')

      InspectMode().do(
        args.fn,
        batch=True if args.mode == 'batch' else False,
        cmdlines=cmdlines,
        abort_on_errors=args.abort_on_errors,
        force_opener=args.force_opener,
      )
    elif args.mode == 'scan':
      if not args.fn:
        ui.fatal('need target')

      from trueseeing.core.exc import InvalidFileFormatError
      try:
        from trueseeing.app.scan import ScanMode
        app = ScanMode(
          args.fn,
          outform=args.scan_report_format,
          outfile=args.scan_output_filename,
          sigsels=args.scan_sigs.split(',') if args.scan_sigs else [],
          excludes=args.scan_exclude_packages if args.scan_exclude_packages else [],
          force_opener=args.force_opener,
        )
        if args.scan_update_cache:
          return self._launch(app.reanalyze())
        else:
          return self._launch(app.scan(oneshot=args.scan_no_cache))
      except InvalidFileFormatError:
        ui.fatal('cannot recognize format')
    else:
      assert False, f'unknown mode: {args.mode}'

def entry() -> None:
  from trueseeing.core.exc import FatalError
  try:
    _require_platform()
    Shell().invoke()
  except FatalError:
    sys.exit(2)

def _require_platform() -> None:
  import sys
  val = sys.platform
  if val == 'win32':
    ui.fatal(f'platform not supported: {val} (consider using containers)')



from __future__ import annotations
from typing import TYPE_CHECKING, NamedTuple

import re
from collections import deque

from trueseeing.core.model.cmd import CommandMixin
from trueseeing.core.ui import ui, OpFormatter, OpLister

if TYPE_CHECKING:
  from typing import Optional
  from trueseeing.api import CommandHelper, Command, CommandMap
  from trueseeing.core.android.context import APKContext

class QualNameInfo(NamedTuple):
  clazz: str
  method: Optional[str]

class ShowCommand(CommandMixin):
  def __init__(self, helper: CommandHelper) -> None:
    self._helper = helper

  @staticmethod
  def create(helper: CommandHelper) -> Command:
    return ShowCommand(helper)

  def get_commands(self) -> CommandMap:
    return {
      'pd':dict(e=self._show_disasm, n='pd qualname', d='show disassembled class/method', t={'apk'}),
      'pk':dict(e=self._show_solved_constant, n='pk[!] 0xaddr index', d='guess and show what constant would flow into the index-th arg of op (!: try harder)', t={'apk'}),
      'pk!':dict(e=self._show_solved_constant, t={'apk'}),
      'pt':dict(e=self._show_solved_typeset, n='pt 0xaddr index', d='guess and show what type would flow into the index-th arg of op', t={'apk'}),
    }

  async def _show_disasm(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    try:
      qn = args.popleft()
    except IndexError:
      ui.fatal('need a qualname')

    context: APKContext = self._helper.get_context().require_type('apk')
    await context.analyze()

    try:
      c = self._parse_qualname(qn)
    except ValueError:
      ui.fatal(f'invalid qualname (try quoting): {qn}')

    with context.store().query().scoped() as q:
      OpLister(OpFormatter(q)).list_tagged(q.body(c.clazz, c.method))

  def _parse_qualname(self, n: str) -> QualNameInfo:
    m = re.fullmatch('(L[^ ]+?;)(->[^ ]+?)?', n)
    if m is None:
      raise ValueError('invalid dalvik name: {n}') # XXX
    return QualNameInfo(
      clazz=m.group(1),
      method=m.group(2)[2:] if m.group(2) else None,
    )

  async def _show_solved_constant(self, args: deque[str]) -> None:
    self._helper.require_target()

    cmd = args.popleft()

    if len(args) < 2:
      ui.fatal('need addr and index')

    addr = int(args.popleft(), 16)
    idx = int(args.popleft())

    limit = self._helper.get_graph_size_limit(self._helper.get_modifiers(args))

    from trueseeing.core.android.analyze.flow import DataFlow
    with DataFlow.apply_max_graph_size(limit):
      context = await self._helper.get_context().require_type('apk').analyze()
      store = context.store()
      q = store.query()
      op = q.op_get(addr)
      if op is not None:
        if cmd.endswith('!'):
          vs = DataFlow(q).solved_possible_constant_data_in_invocation(op, idx)
          ui.info(repr(vs))
        else:
          try:
            v = DataFlow(q).solved_constant_data_in_invocation(op, idx)
            ui.info(repr(v))
          except DataFlow.UnsolvableValueError as ce:
            if not ui.is_debugging:
              ui.error('value is not solvable (possibly not a compile-time constant); try enabling debug mode (e core.debug=true) to see the data graph')
            else:
              ui.error(f'value is not solvable (possibly not a compile-time constant), data graph: {ce.graph}')
          except DataFlow.NoSuchValueError as e:
            ui.error(str(e))
      else:
        ui.error('0x{:08x}: invalid address'.format(addr))

  async def _show_solved_typeset(self, args: deque[str]) -> None:
    self._helper.require_target()

    _ = args.popleft()

    if len(args) < 2:
      ui.fatal('need addr and index')

    addr = int(args.popleft(), 16)
    idx = int(args.popleft())

    limit = self._helper.get_graph_size_limit(self._helper.get_modifiers(args))

    from trueseeing.core.android.analyze.flow import DataFlow
    with DataFlow.apply_max_graph_size(limit):
      context = await self._helper.get_context().require_type('apk').analyze()
      store = context.store()
      q = store.query()
      op = q.op_get(addr)
      if op is not None:
        vs = DataFlow(q).solved_typeset_in_invocation(op, idx)
        ui.info(repr(vs))
      else:
        ui.error('0x{:08x}: invalid address'.format(addr))



from __future__ import annotations
from typing import TYPE_CHECKING

import os
import os.path

from functools import cache
from trueseeing.core.ui import ui
from trueseeing.core.env import get_extension_dir, get_extension_dir_v0, get_extension_package_prefix

if TYPE_CHECKING:
  from typing import Any, Dict, ClassVar, Optional, Iterable, Iterator, Type, TypeVar
  from typing_extensions import Final
  from trueseeing.api import Command, Signature, FileFormatHandler

  T = TypeVar('T')

class Extension:
  _ns: Any
  _inst: ClassVar[Optional[Extension]] = None
  _module_name: Final[str] = 'ext'
  disabled: ClassVar[bool] = False

  @classmethod
  def get(cls) -> Extension:
    if cls._inst is None:
      cls._inst = Extension()
    return cls._inst

  def __init__(self) -> None:
    self._ns = {}
    if not self.disabled:
      self._ns.update(self._import())
      self._ns.update(self._compile())

  def _import(self) -> Any:
    import sys
    from glob import iglob
    from importlib import import_module
    from importlib_metadata import packages_distributions

    o = {}
    prefix = get_extension_package_prefix()
    for n in packages_distributions():
      if not n.startswith(prefix):
        continue
      o[n] = import_module(n)
    # We need to discover *.pth by our own
    for p in sys.path:
      if os.path.isfile(p):
        continue
      for path in iglob('{}/{}*.pth'.format(p, prefix)):
        n = os.path.basename(path).replace('.pth', '')
        o[n] = import_module(n)
    return o

  def _compile(self) -> Any:
    globals_: Dict[str, Any] = dict(__name__='__main__', ui=ui)
    locals_: Dict[str, Any] = dict()
    try:
      starter: Optional[str] = None
      path = get_extension_dir()
      if os.path.isdir(path):
        starter = self._importer(path)
      else:
        for trypath in self._get_extensions_v0():
          if os.path.exists(trypath):
            ui.warn('extension uses old path ({vers}), consider moving it to {path}'.format(vers='v0', path=os.path.join(get_extension_dir(), os.path.basename(trypath))))
            starter = self._importer_v0(trypath)
            break
      if starter is not None:
        code = compile(starter, filename='<string>', mode='exec')
        exec(code, globals_, locals_)
        return locals_
      else:
        return {}
    except Exception as e:
      ui.warn('Uncaught exception during invocation', exc=e)
      return {}

  def get_signatures(self) -> Iterator[Type[Signature]]:
    from trueseeing.api import Signature
    from trueseeing.core.tools import get_public_subclasses, get_missing_methods
    for _, m in self._ns.items():
      for clazz in get_public_subclasses(m, Signature):  # type: ignore[type-abstract]
        missing = get_missing_methods(clazz)
        if missing:
          from trueseeing.core.tools import get_fully_qualified_classname
          ui.warn('ignoring signature {}: missing methods: {}'.format(get_fully_qualified_classname(clazz), ', '.join(missing)))
          continue
        yield clazz

  @cache
  def get_commands(self) -> Iterator[Type[Command]]:
    from trueseeing.api import Command
    from trueseeing.core.tools import get_public_subclasses, get_missing_methods
    for _, m in self._ns.items():
      for clazz in get_public_subclasses(m, Command):  # type: ignore[type-abstract]
        missing = get_missing_methods(clazz)
        if missing:
          from trueseeing.core.tools import get_fully_qualified_classname
          ui.warn('ignoring command {}: missing methods: {}'.format(get_fully_qualified_classname(clazz), ', '.join(missing)))
          continue
        yield clazz

  def get_fileformathandlers(self) -> Iterator[Type[FileFormatHandler]]:
    from trueseeing.api import FileFormatHandler
    from trueseeing.core.tools import get_public_subclasses, get_missing_methods
    for _, m in self._ns.items():
      for clazz in get_public_subclasses(m, FileFormatHandler):  # type: ignore[type-abstract]
        missing = get_missing_methods(clazz)
        if missing:
          from trueseeing.core.tools import get_fully_qualified_classname
          ui.warn('ignoring file format handler  {}: missing methods: {}'.format(get_fully_qualified_classname(clazz), ', '.join(missing)))
          continue
        yield clazz

  # XXX: gross hack
  def _importer(self, path: str, *, only: Optional[str] = None) -> Optional[str]:
    from glob import iglob
    import re
    path = os.path.expandvars(os.path.expanduser(path))
    mods: Dict[str, str] = dict()

    def _get_seeds() -> Iterator[str]:
      if only:
        yield os.path.join(path, only)
      else:
        for n in iglob(f'{path}/*'):
          yield n

    def _discover() -> Iterator[str]:
      for n in _get_seeds():
        bn = os.path.basename(n)
        if os.path.isfile(n):
          ns, ne = os.path.splitext(bn)
          if not (ne in ['py', 'pyc', 'pyo']):
            ui.warn(f'cannot load extension {bn}: invalid filename')
            continue
          if not re.fullmatch(r'[0-9A-Za-z_]+', ns):
            ui.warn(f'cannot load extension {bn}: invalid filename')
            continue
          yield n
        elif os.path.isdir(n):
          if os.path.exists(os.path.join(n, '__init__.py')):
            yield n
            continue
          found = False
          for nn in iglob(f'{n}/*'):
            if os.path.exists(os.path.join(nn, '__init__.py')):
              yield nn
              found = True
          if not found:
            ui.warn(f'cannot load extension {bn}: need to be or contain a module')
        else:
          ui.warn(f'cannot load extension {bn}: invalid filetype')

    for n in _discover():
      ns, _ = os.path.splitext(os.path.basename(n))
      mods[ns] = n

    return 'import sys\ntry:\n sys.dont_write_bytecode=True;\n{importblock}\nfinally:\n sys.dont_write_bytecode=False;del sys'.format(
      importblock='\n'.join([
        ' try:\n'
        '  sys.path[0:0]=["{path}"]\n'
        '  import {ns}\n'
        ' except Exception as e:\n'
        '  ui.warn(f"cannot load extension {bn}: {{e}}")\n'
        ' finally:\n'
        '  del sys.path[0]\n'.format(ns=ns, bn=os.path.basename(path), path=os.path.dirname(path)) for ns, path in mods.items()
      ])
    )

  def _importer_v0(self, path: str) -> Optional[str]:
    root = get_extension_dir_v0()
    return self._importer(root, only=os.path.relpath(path, start=root))

  def _get_extensions_v0(self) -> Iterable[str]:
    for fn in 'ext.pyc', 'ext.py', 'ext':
      yield os.path.join(get_extension_dir_v0(), fn)
